/*
 *    fgrab.c  -- Generic Frame Grabber.
 *    
 *    See file COPYING for copyright info
 *
 */

/*
 * HISTORY
 * =======
 *
 *
 *
 */


/*
 * fgrab.c contains the most generic code of the frame grabbers.  That means
 *      1) Interfaces to the OS like read, write, etc calls
 *      2) Buffer and irq allocation and freeing
 *      3) Passing buffer allocations back to user
 *      4) Controlling when DMA and interrupts may occur
 *      5) Counting frames and determining which frame buffer is current.
 * It leaves to lower level routines the following
 *      1) Layout of data within a frame
 *      2) Selection of input device and data format
 * It depends on lower level routines to enable and disable interrupts and DMA
 */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <asm/irq.h>
#include <asm/pgtable.h>
#include <linux/time.h>
#include <linux/malloc.h>
#include <linux/bigphysarea.h>

#include "fgrab_ioctl.h"
#define FGRAB_INTERNAL
#include "fgrab.h"
#undef  FGRAB_INTERNAL

/*
 * Globals:
 */


static u_short FgrabDevs;
static struct fgrab *fgrabs[MAXFGRABS];
static unsigned intr_status;    /* used to synchronize processing of slow IRQ   */

/*
 * Prototypes:
 */
int  fgrab_register(int irq,                    /* desired irq              */
                    struct fgrab_progs *progs,  /* board level routines     */
                    unsigned int d_size,        /* size of data area        */
                    void **d_ptr,               /* place to return pointer  */
                    unsigned int *dev) ;        /* minor device             */
void fgrab_unregister(unsigned minor) ;
void fgrab_wait_jiffies(u_long waitJiffies) ;
static void fgrab_setup(void) ;
static struct fgrab *get_fgb_from_inode(struct inode *ip);
static void free_frame_buffer(struct fgrab *fgb);
static int alloc_frame_buffer(struct fgrab *fgb);
static void init_per_frame(struct fgrab_per_frame *pf) ;
static int read_fs(const void *useraddr, void *where, u_int size);
static int write_fs_val(const u_long value, void * useraddr, u_int size);
static int write_fs_ptr(const void * ptr, void * useraddr, u_int size);
static void fgrab_unwind(void);
static void fgrab_intr(int irq, void *dev_id, struct pt_regs *fake);
static void fgrab_intr_proc(struct fgrab *fgb);
static void abort_intr(struct fgrab *fgb, struct timeval ts) ;
static void start_capture(struct fgrab *fgb, unsigned type);
static void stop_capture(struct fgrab *);
static int bsd_fgrab_read(struct fgrab *fgb, char *user_buf, int user_count);
static int bsd_fgrab_ioctl(struct fgrab *fgb, unsigned int cmd, void *arg);


#ifdef    FGRAB_EVENT_LOG
#define FGRAB_EVENT_LOG_NULL_    0
#define FGRAB_EVENT_LOG_INTR_    1
#define         FGRAB_EVENT_LOG_INTR(data) record_event(FGRAB_EVENT_LOG_INTR_,data) ;
#define FGRAB_EVENT_LOG_PROC_    2
#define         FGRAB_EVENT_LOG_PROC(data) record_event(FGRAB_EVENT_LOG_PROC_,data) ;
#define FGRAB_EVENT_LOG_EPROC_   3
#define         FGRAB_EVENT_LOG_EPROC(data) record_event(FGRAB_EVENT_LOG_EPROC_,data) ;
#define FGRAB_EVENT_LOG_INIT init_event_log();
#define FGRAB_EVENT_LOG_STOP stop_record();
#define FGRAB_EVENT_LOG_PRINT print_event_log();
static void init_event_log(void) ;
static void record_event(unsigned event, unsigned data) ;
static void stop_record(void);
static void print_event_log(void) ;
#else  /* FGRAB_EVENT_LOG */
#define         FGRAB_EVENT_LOG_INTR(data) 
#define         FGRAB_EVENT_LOG_PROC(data)
#define         FGRAB_EVENT_LOG_EPROC(data)
#define FGRAB_EVENT_LOG_INIT
#define FGRAB_EVENT_LOG_STOP
#define FGRAB_EVENT_LOG_PRINT
#endif /* FGRAB_EVENT_LOG */


static struct symbol_table fgrab_syms = {
#include <linux/symtab_begin.h>
    X(fgrab_register),
    X(fgrab_unregister),
    X(fgrab_wait_jiffies),
#include <linux/symtab_end.h>
};

static struct fgrab *get_fgb_from_inode(struct inode *ip) {
    int minor = MINOR(ip->i_rdev);
    if (minor >= FgrabDevs)	/* unit out of range */
        return(0);
    return( fgrabs[minor] );
}

static int read_fs(const void *useraddr, void *where, u_int size) {
    int err;

    if (useraddr == NULL || !size)
        return -EINVAL;
    if ((err = verify_area(VERIFY_READ, useraddr, size)))
        return err;
    switch (size) {
    case 1:
        *(u_char *)where = get_user((u_char *) useraddr);
        break;
    case 2:
        *(u_short *)where = get_user((u_short *) useraddr);
        break;
    case 4:
        *(u_long *)where = get_user((u_long *) useraddr);
        break;
    default:
        memcpy_fromfs(where, useraddr, size);
    }
    return 0;
}

static int write_fs_val(const u_long value, void * useraddr, u_int size) {
   int err;

   if (useraddr == NULL)
       return -EINVAL;
   if ((err = verify_area(VERIFY_WRITE, useraddr, size)))
       return err;
   switch (size) {
   case 1:
       put_user((u_char)value, (u_char *) useraddr);
       break;
   case 2:
       put_user((u_short)value, (u_short *) useraddr);
       break;
   case 4:
       put_user((u_long)value, (u_long *) useraddr);
       break;
   default:
       return -EINVAL;
   }
   return 0;
}

static int write_fs_ptr(const void * ptr, void * useraddr, u_int size) {
    int err;
    
    if (useraddr == NULL)
        return -EINVAL;
    if ((err = verify_area(VERIFY_WRITE, useraddr, size)))
        return err;
    memcpy_tofs(useraddr, ptr, size );
    return 0;
}

static void fgrab_setup(void) {
    int i;
    for (i=0;i<MAXFGRABS;i++) fgrabs[i] = NULL ;
    intr_status = 0;
}

/* fgrab_register is called by lower level board specific drivers when
   they detect a board.  It assigns a minor device number, allocates
   an fgrab structure and the data space they request.  The address of
   the data space and the minor device number (for kernel messages)
   are passed back.                                                         */
int  fgrab_register(int irq,                    /* desired irq              */
                    struct fgrab_progs *progs,  /* board level routines     */
                    unsigned int d_size,        /* size of data area        */
                    void **d_ptr,               /* place to return pointer  */
                    unsigned int *dev) {        /* minor device             */
    struct fgrab *fgb;
    void *mptr;
    unsigned int minor;
    /* check for any released minor devices                                 */
    for (minor=0;minor<FgrabDevs;minor++)
        if (fgrabs[minor]==0) break;
    /* if no released minor devices then move on                            */
    if (minor==FgrabDevs) {
        if (minor >= MAXFGRABS) {
            printk("fgrab: Maximum number of frame grabbers exceeded\n");
            return(-1);
        }
        fgrabs[minor] = NULL ;  /* initialize to empty in case we bail out  */
        FgrabDevs++;
    }
    *dev = minor;
    mptr = kmalloc(sizeof(struct fgrab)+d_size,GFP_KERNEL);
    if ( ! mptr ) {
        printk("fgrab: Cannot allocate memory for minor device %d\n",minor);
        return(-1);
    }
    memset(mptr, 0, sizeof(struct fgrab)+d_size); /* Clear/NULL everything */
    fgb = (struct fgrab *) mptr;
    *d_ptr = mptr + sizeof(struct fgrab) ;
    *dev = minor;
    fgrabs[minor] = fgb ;
    fgb->minor = minor ;
    fgb->data = *d_ptr ;
    fgb->progs = progs ;

    /* Request a slow sharable interrupt */
    if ( request_irq(irq, fgrab_intr, (SA_SHIRQ), "fgrab", fgb) ) {
        printk("fgrab: cannot allocate irq %d\n",irq);
        fgrabs[minor] = NULL ;
        kfree(mptr);
        return(-1);
    }
    fgb->irq = irq;
    fgb->flags |= FGRAB_INITALIZED ;
    PRINTK((KERN_INFO "fgrab%d: registered with irq %d\n",minor,irq));

    return(0);
}

void fgrab_unregister(unsigned minor) {
    struct fgrab *fgb = fgrabs[minor];
    if (fgb == NULL) return;
    fgrabs[minor] = NULL ;
    free_irq(fgb->irq, fgb);
    free_frame_buffer(fgb);
    kfree(fgb);         /* Unwind kalloc() */
}

/*
 * Delay for WaitJiffies via timeout and schedule().
 * We need to be technically interruptable for this to work,
 * but we really don't want to be, so we block everything.
 * Don't call from interrupt.
 */
void fgrab_wait_jiffies(u_long waitJiffies)
{
  u_long saveBlocked = current->blocked;
  u_long endJiffies = jiffies + waitJiffies;

  current->blocked = ~0; /* Block everything we can, ignore the rest */
  while( jiffies < endJiffies ) {
    current->timeout = endJiffies;
    current->state = TASK_INTERRUPTIBLE;
    schedule();
  }
  current->blocked = saveBlocked;
}


/* This 4 meg characteristic comes from the SAA7116.  It may be appropriate
   to relax it in the future						    */
#define FOURMEG (4 * 1024 * 1024)
static int alloc_frame_buffer(struct fgrab *fgb) {
    u_long startAddr, block_size_left, left_over;
    long bufSize;
    int frame_size, i;
    
    /* Things have changed here.  There may be some backwards incompatibility
     * but I hope not too much.
     *
     * It is important that frames do not cross 4Mb boundaries.
     * The dma_end registers only have 22 significant bits, and even
     * without range checking the dma regs themselves don't seem to
     * increment over 4Mb boundaries (they just wrap around into 
     * the current 4Mb block).
     */

    if ( fgb->frame_buffer ) {
        free_frame_buffer(fgb);
    }
    /* 
     * time struct added with a frame ?? 
     */
    frame_size = fgb->frame_size;
    /* following was conditioned on (fgb->flags & FGRAB_WANT_TS) --
     * do it always to avoid possibility of SET_TS after SET_GEO scribbling
     * on the following frame or causing access to unallocated core.
     * It would be better to move this into the fgrab_mem structure. RWF 3JUN97
     */
    frame_size += sizeof(struct fgrab_per_frame);

    /* 
     * Iterate until all the frames fit in without crossing a 4Mb boundary 
     */
    fgb->frame_buffer = NULL; left_over = 0;
    while (!fgb->frame_buffer) {
        PRINTK((KERN_DEBUG "fgrab%d: Trying to allocate frames\n",fgb->minor));
        bufSize = fgb->frames*frame_size + left_over;
        /* take leftover up to next page boundary       */
        left_over += PAGE_ALIGN(bufSize)-bufSize;
        bufSize = PAGE_ALIGN(bufSize);
        if ( ! (startAddr = (u_long)bigphysarea_alloc(bufSize + PAGE_SIZE)) ) {
            PRINTK((KERN_INFO "fgrab%d: frame buffer failed  %ld\n",
                    fgb->minor, bufSize));
            return -ENOMEM;
        }
        fgb->frame_buffer = (void *)startAddr;
        fgb->mem = fgb->frame_buffer + bufSize;
        fgb->fb_size = bufSize + PAGE_SIZE;
        PRINTK((KERN_DEBUG "fgrab%d: frame_buffer=%lx, mem=%lx, fb_size=%x\n", 
                fgb->minor, startAddr, (u_long)fgb->mem, fgb->fb_size));

        /* 
         * layout frames in memory 
         */
        i = 0;
        while (i<fgb->frames) {
            block_size_left = FOURMEG - ((FOURMEG-1) & startAddr);
            if (block_size_left > frame_size) {
                fgb->frame_offset[i] = startAddr - (u_long)fgb->frame_buffer;
                PRINTK((KERN_DEBUG "fgrab%d: frame %d: %lx -> %lx\n", fgb->minor, i,
                        (u_long)fgb->frame_buffer+fgb->frame_offset[i],
                        (u_long)fgb->frame_buffer+fgb->frame_offset[i]+frame_size));
                bufSize -= frame_size;
                startAddr += frame_size;
                i++;
            } else {
                bufSize -= block_size_left;
                startAddr += block_size_left;
            }
        }
        if (bufSize < 0) {
            PRINTK((KERN_DEBUG "fgrab%d: left_over=0x%x,%d bufSize=%d -bufSize=0x%x\n",
               fgb->minor, (unsigned)left_over,(unsigned)left_over,(int)bufSize,(unsigned)(-bufSize)));
            left_over += -bufSize;
            bigphysarea_free(fgb->frame_buffer, fgb->fb_size);
            fgb->frame_buffer = 0;
        }
    }
    /* done */  
    
    PRINTK((KERN_DEBUG "fgrab%d: Alloc_frame_buffer Done\n",fgb->minor));
    return 0;
}


static void free_frame_buffer(struct fgrab *fgb) {
    if ( fgb->frame_buffer ) {
        PRINTK(("fgrab%d: freeing frame buffer address 0x%x\n",
                fgb->minor, (unsigned)fgb->frame_buffer));
        bigphysarea_free(fgb->frame_buffer, fgb->fb_size);
    }
    fgb->frame_buffer = NULL;
}

static int fgrab_open(struct inode * inode, struct file * file) {
    struct fgrab *fgb;

    if ( !(fgb = get_fgb_from_inode(inode)) )
        return(-ENXIO);
    if (!(fgb->flags & FGRAB_INITALIZED))	/* device not found */
        return(-ENXIO);
    if (fgb->flags & FGRAB_OPEN)		/* device is busy */
        return(-EBUSY);
    MOD_INC_USE_COUNT;
    FGRAB_EVENT_LOG_INIT
    fgb->flags |= FGRAB_OPEN;

    fgb->progs->open(fgb->minor);
    
    fgb->fifo_errors = 0;
    fgb->dma_errors = 0;
    fgb->bus_errors = 0;
    fgb->frames_captured = 0;
    fgb->even_fields_captured = 0;
    fgb->odd_fields_captured = 0;
    fgb->proc = (struct task_struct *)0;

    return 0;
}

static void fgrab_close(struct inode * inode, struct file * file) {
    struct fgrab *fgb;
    
    if ( (fgb = get_fgb_from_inode(inode)) ) {
        FGRAB_EVENT_LOG_PRINT;
        if ( (fgb->fifo_errors!=0) || (fgb->dma_errors!=0) || (fgb->bus_errors!=0) ) {
            PRINTK(("fgrab%d: ERRORS Fifo: %lu  Dma: %lu  Bus: %lu\n", 
                    fgb->minor, fgb->fifo_errors, fgb->dma_errors, fgb->bus_errors));
        }
        PRINTK(("fgrab%d: (fgb 0x%x) close.\n", fgb->minor, (unsigned)fgb));
        fgb->flags &= ~FGRAB_OPEN;
        stop_capture(fgb);
        fgb->flags &= ~FGRAB_ENABLE_INTR ;
        fgb->progs->disable_intr(fgb->minor);
        fgb->progs->disable_dma(fgb->minor);
        free_frame_buffer(fgb);
        fgb->progs->close(fgb->minor);
        MOD_DEC_USE_COUNT;
    }
}

static int fgrab_read(struct inode * inode, struct file * file, 
                      char * buffer, int count) {
    struct fgrab *fgb;
    int ret;
    
    if ( !(fgb = get_fgb_from_inode(inode)) )
        return(-ENXIO);
    ret = bsd_fgrab_read(fgb, buffer, count);
    return ret;
}

static int bsd_fgrab_read(struct fgrab *fgb, char *user_buf, int user_count) {
    u_int fgrab_count = (fgb->rows * fgb->cols * fgb->depth) >> 3;

    /* no frame buffer allocated (ioctl failed) */
    if (fgb->frame_buffer == NULL)
        return(-ENOMEM);

    if (fgb->flags & FGRAB_CAP_MASK)
        return(-EIO);		/* already capturing */

    if ( user_count != fgrab_count )
        return(-EINVAL);

    start_capture(fgb, FGRAB_SINGLE);
    interruptible_sleep_on(&fgb->waitq);
    if (current->signal & ~current->blocked) {
        stop_capture(fgb);
        return -EINTR;
    }
    /* NOTE: added +fgb->framne_offset[0] 17/7/96 */
    memcpy_tofs(user_buf,fgb->frame_buffer+fgb->frame_offset[0],fgrab_count);
    fgb->flags &= ~(FGRAB_SINGLE | FGRAB_WANT_MASK);

    return(user_count);
}




/*
 * Writes are meaningless.
 */
static int fgrab_write(struct inode * inode, struct file * file,
			const char * buffer, int count) {
    return -EPERM;
}

static int fgrab_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg) {
    struct fgrab *fgb;
    int ret;
    if ( !(fgb = get_fgb_from_inode(inode)) )
        return(-ENXIO);
    ret = bsd_fgrab_ioctl(fgb, cmd, (void *)arg);
    return ret;
}

static int fgrab_mmap(struct inode * inode, struct file * file,
		    struct vm_area_struct * vma) {
    struct fgrab *fgb;
    
    int err;
    u_long size = vma->vm_end - vma->vm_start;
    u_long offset;
    
    if ( !(fgb = get_fgb_from_inode(inode)) )
        return -ENXIO;
    offset = (u_long)fgb->frame_buffer;
    if ( (offset|size) & ~PAGE_MASK )
        return -EFAULT;
    if ( size > fgb->fb_size ) {
        return -EFBIG;
    }
    /*
     * changed based on /usr/src/linux/kernel/mem.c 
     * (Darius.Burschka@lpr.e-technik.tu-muenchen.de)
     */
    vma->vm_offset=offset;
    if(vma->vm_offset >=high_memory)
        pgprot_val(vma->vm_page_prot) |= _PAGE_PCD;
    if ( (err=remap_page_range(vma->vm_start, offset,
                               size, vma->vm_page_prot)) )
        return err;
    vma->vm_inode = inode;
    inode->i_count++;
    return 0;
}

static void fgrab_unwind(void) {
    struct fgrab *fgb;
    int i;
    unregister_chrdev(FGRAB_MAJOR, "fgrab");
    for ( i = 0; i < MAXFGRABS; i++ ) {
        if ((fgb=fgrabs[i]) != NULL) {
            fgb->progs->unwind(fgb->minor);
        }
    }   	    
}




static struct file_operations fgrab_fops = {
        NULL,		/* seek */
	fgrab_read,	/* read */
	fgrab_write,	/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	fgrab_ioctl,  	/* ioctl */
	fgrab_mmap,   	/* mmap */
        fgrab_open,     /* open */
        fgrab_close,	/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL		/* revalidate */
};


/* Designed to be used as a module */
#ifdef MODULE

void cleanup_module(void) {
    fgrab_unwind();
}


int init_module(void) {
    register_symtab(&fgrab_syms);

    if (register_chrdev(FGRAB_MAJOR,"fgrab",&fgrab_fops)) {
        PRINTK((KERN_INFO "fgrab: unable to get major %d for fgrab devices\n",
                FGRAB_MAJOR));
        return -EIO;
    }
    fgrab_setup();
    return(0);
}
#endif /*MODULE*/


static void
start_capture(struct fgrab *fgb, unsigned type) {
    u_long buf_e, buf_o;
    struct fgrab_per_frame *pf;
    int i;
    
    fgb->progs->clear_err(fgb->minor);  /* Clear errors         */
    fgb->synch_wait = 0 ;
    fgb->flags |= type;
    fgb->cur_frame = 1;
    buf_e = (u_long)(fgb->frame_buffer + fgb->frame_offset[0]);
    /*
     * If more than one frame and we are not interlaced then need to put even
     * and odd fields on different buffers
     */
    if ( ( (type&FGRAB_SYNCAP) || (type&FGRAB_ONCE) )                   && 
         ( (fgb->flags&FGRAB_FIELDS_MASK) != FGRAB_INTERLACE_FIELDS )   &&
         ( fgb->frames > 1 )                                            ) {
        buf_o = (u_long)(fgb->frame_buffer + fgb->frame_offset[1]);
        fgb->next_frame = 2 ;
    } else {
        buf_o = buf_e ;
        fgb->next_frame = 1 ;
    }    
    cli();      /* check and set of dma must be atomic wrt interrupts RWF 13aug97       */
    fgb->progs->set_dma_e(fgb->minor,
                          fgb->dma_add_e, fgb->dma_str_e,
                          buf_e, fgb->frame_size);
    fgb->progs->set_dma_o(fgb->minor,
                          fgb->dma_add_o, fgb->dma_str_o,
                          buf_o, fgb->frame_size);
    sti();
    fgb->progs->enable_dma(fgb->minor);
    for (i=0;i< ((fgb->flags & FGRAB_SYNCAP) ? fgb->frames : 1) ; i++) {
        pf = (struct fgrab_per_frame *)(fgb->frame_size + (u_long)(fgb->frame_buffer) + fgb->frame_offset[i]);
        init_per_frame(pf);
    }
    
    switch(fgb->flags & FGRAB_FIELDS_MASK) {
    case FGRAB_ONLY_EVEN_FIELDS:
        fgb->flags |= FGRAB_WANT_EVEN;
        fgb->flags &= ~FGRAB_EXPECT_FIELD;
        break;
    case FGRAB_ONLY_ODD_FIELDS:
        fgb->flags |= FGRAB_WANT_ODD;
        fgb->flags |= FGRAB_EXPECT_FIELD;
        break;
    default:
        fgb->flags |= FGRAB_WANT_MASK;
        fgb->flags &= ~FGRAB_EXPECT_FIELD;
        break;
    }
    fgb->progs->start_capture(fgb->minor,(type==FGRAB_SINGLE)?1:0);
    fgb->flags |= FGRAB_ENABLE_INTR ;
    fgb->progs->enable_intr(fgb->minor);
}

static void
stop_capture(struct fgrab *fgb) { /* Stop the capture */
    fgb->progs->stop_capture(fgb->minor);
    fgb->flags &= ~(FGRAB_CAP_MASK|FGRAB_WANT_MASK);
}

/* interrupt handling routine 
 *  complete fgrab_read() if using interrupts
 * fgrab_intr handles the synchronination needed for a slow shared IRQ
 * The actual processing of boards that need attention is in fgrab_intr_proc.
 * This keeps all the cli-sti pairs in fgrab_intr and removes one level of
 * indentation from the code in fgrab_intr_proc.
 */

static void
fgrab_intr(int irq, void *dev_id, struct pt_regs *fake) {
    int dev;
    struct fgrab *fgb;
    FGRAB_EVENT_LOG_INTR(intr_status);
    /* Flag that we need a pass through the loop                */
    cli();
    intr_status |= FGRAB_INTR_NEED_LOOP ;
    if (intr_status & FGRAB_INTR_PROC_ACTIVE) {
        /* if another interrupt is active then just return      */
        sti();
        return;
    }
    /* Need to run interrupt processing loop                    */
    /* First mark that we are going to run it                   */
    intr_status |= FGRAB_INTR_PROC_ACTIVE ;
    /* Start the loop that goes till all interrupts are handled */
    while(intr_status & FGRAB_INTR_NEED_LOOP) {
        /* Start this loop with no other pass needed            */
        intr_status &= ~FGRAB_INTR_NEED_LOOP ;
        sti();
        /*
         * More than one device could be using this interrupt, so ignore
         * irq and dev_id and check all devices.
         */
        for ( dev = 0; dev < FgrabDevs; dev++ ) {
            if ( (fgb = fgrabs[dev]) == NULL) 
                continue;       /* no board here                        */
            if ( (fgb->flags&FGRAB_ENABLE_INTR) == 0) 
                continue;       /* interrupts not enabled here          */
            if (fgb->progs->check_intr(fgb->minor))
                continue;       /* This board does not want attention   */
            fgrab_intr_proc(fgb);
        }
        cli();  /* block interrupts while we check loop flag    */
    }
    /* Mark that we have exited interrupt processing loop       */
    intr_status &= ~FGRAB_INTR_PROC_ACTIVE ;
    sti();
    return;
}

    /*
     * Notes on the interpretation of SYNCAP, CONTIN, ONCE, and SINGLE flags:
     *   SYNCAP and (CONTIN|ONCE)  are modes where sequential frames are captured
     *   into sequential buffers.  Both modes set the flags and counts telling which
     *   buffers have data.  In SYNCAP mode the user resets the flags or decrements
     *   the counter telling the driver that it is through with the data in a buffer.
     *   The driver will not over-write data it has indicated is present until
     *   the user indicates it is done with the data.  Frames which come in
     *   with no place to go will be dropped on the floor.  In (CONTIN|ONCE) mode the 
     *   driver indicates that it has finished with a frame by setting its flag
     *   and incrementing the counter.  When all buffers have been filled the
     *   capture stops. CONTIN (without ONCE) and SINGLE both write to a single
     *   buffer.  CONTIN continuously updates the buffer as new data comes in
     *   while SINGLE causes a single frame to be captured.
     * Notes on the synchronization methods:
     *   The original meteor driver used the num_active_bufs variable in the
     *   shared memory structure to do the handshaking.  That mode is preserved
     *   in this driver primarily for compatibility.  It has been suplemented
     *   by a second handshaking mode which uses a flag word per frame to do
     *   the hand shaking.  The reason for the new method is that the read-modify-write
     *   operation performed by the user code is not quaranteed to be atomic.  That
     *   is it is possible for the board to generate an interupt that occurs after
     *   the user code has done the read but before it has done the write.
     *   Then if the driver changes that memory location the net result is that
     *   the driver's change will be overwritten by the user mode code when it
     *   restarts.  This is a low probability event but one that has the potential
     *   for creating a real nuisance for systems which are run continuously for
     *   long periods of time.  The old synchronous mode was enabled by setting
     *   either lowat or hiwat to a positive number less than the number of frames.  
     *   The new mode is enabled by setting lowat=hiwat=-1. The old mode stopped
     *   when num_active_bufs exceeded hiwat and started again when num_active_bufs
     *   went below lowat.  The new mode stops only when there are no available
     *   buffers and restarts as soon as a buffer becomes available.  The flag word
     *   is part of the per-frame structure that is at the end of the frame buffer.
     *   This structure includes the time stamp that used to appear at the end of
     *   the buffer as its first element but has the new flag word as well as 
     *   an event log of the time of various interrupts associated with this frame.
     * Notes on interrupt routine changes:
     *   Adding in the new synchronization method required taking a detailed look
     *   at the interrupt routine.  It appeared that the decision on whether or
     *   not to advance to the next frame in the synchronized mode was made at the
     *   end of the previous interrupt.  The first thing the interrupt routine did
     *   was to set up the dma registers for the next frame based on the value
     *   of synch_wait determined at the end of the previous interrupt.  Withing the
     *   context of a low water / high water scheme the exact time at which the
     *   decision is made is not critical.  However, the new scheme is supposed to
     *   stop only when it has no option to go on.  Hence the time at which the
     *   decision is made is critical.  Hence the point of decision needed to be
     *   moved.  It also appeared that the advancement to the next buffer was
     *   made after the dma registers had been set up for the next field.  Hence
     *   data would be written into a buffer after the user had been told the
     *   buffer had been completely filled.  This could also explain the phenomena
     *   observed when skipping frames (requesting a frame rate that is a fraction
     *   of the camera rate) where things moving in the frame were quite blurred.
     *   What seems to have been happening is that the even field from one contiguous
     *   even-odd pair was combined with the odd field from another contiguous
     *   even-odd pair which occurred at a later time.  These observations lead to
     *   a reorganization of the interrupt routine.  One consequence of that is the
     *   placement of the setup of dma registers further on in the routine so that
     *   there will be a longer interval between when the interrupt for field done
     *   occurs and the dma registers are set up for the next frame.  I don't
     *   have any documentation on how big that window is -- it may depend on the
     *   number of lines that are not digitized before and after the start of
     *   field.  If it is a fraction of line width then it is a really critical
     *   thing and will probably create a problem.  If that is the case then a small
     *   part of the interrupt routine should be triggered by the linux rt code
     *   to make sure that no other device causes intermittent problems.
     * Notes on change of IRQ from fast to slow:
     *   The meteor driver used a fast IRQ to process its interrupts.  Fast IRQ's
     *   have the characteristic of minimum latency fron the interrupt to the handler,
     *   running the handler with all interrupts disabled, a minimal stack, and not 
     *   triggering a run of the scheduler upon completion.  They are intended for 
     *   devices that generate a high rate of interrupts each of which requires 
     *   minimal processing.  The small stack and not running the scheduler are both
     *   strategies that minimize the latency and overhead associated with these
     *   interrupts.  A prime usage for them is for serial ports which can generate
     *   thousands of interrupts per second.  They really are not appropriate for
     *   devices which generate only tens of interrupts per second - especially if
     *   latency in starting up the associated user process is an issue.  This driver
     *   uses a slow IRQ.  It results in dramatically reduced latency in starting up
     *   user processes runing with posix real-time priorities as well as reducing
     *   the amount of time interrupts are disabled.  An additional side effect is
     *   that a larger stack is available so that heroic measures to limit stack usage
     *   in subroutines are not necessary.  The interrupt routine is now run with
     *   interupts enabled so it is possible that the board will generate an
     *   interrupt while we are processing a previous one.  In addition it is
     *   possible that with more than one board two boards will share an interrupt
     *   and hence one board may interrupt while the other one is being processed.
     *   This has lead to a two-tiered handling of interrupts.  First as each interrupt
     *   comes in it is effectively put on a queue by setting a flag to trigger a
     *   review of all boards to see if any need processing when the current processing
     *   is finished.  Then if another interrupt has not started up the real processing
     *   code it is started up.  Making this work requires judicious use of cli/sti
     *   pairs.
     *          R. Fries 25JUN97
     */
static void
fgrab_intr_proc(struct fgrab *fgb){
    u_long status, this_frame;
    struct fgrab_per_frame *this_pf;
    struct timeval ts;
    struct fgrab_mem *mm = fgb->mem;
    /*
     * Disable future interrupts if a capture mode is not selected.
     * This can happen when we are in the process of closing or 
     * changing capture modes, otherwise it shouldn't happen.
     */
    if(!(fgb->flags & FGRAB_CAP_MASK)) {
        fgb->flags &= ~FGRAB_ENABLE_INTR ;
        fgb->progs->disable_intr(fgb->minor);
    }
    /*
     * set up pointers to this frame to record status
     */
    this_frame = fgb->cur_frame-1 ;
    this_pf = (struct fgrab_per_frame *)(fgb->frame_size + 
                   (u_long)fgb->frame_buffer + fgb->frame_offset[this_frame]) ;
    /*
     * Check for errors and return status telling which frame.
     */
    status = fgb->progs->check_err(fgb->minor);
    FGRAB_EVENT_LOG_PROC(status);
    if (status & FGRAB_STATUS_FIFO_ERR) {
        this_pf->error |= FGRAB_ERROR_FIFO;
        fgb->fifo_errors++;
    }
    if (status & FGRAB_STATUS_DMA_ERR) {
        this_pf->error |= FGRAB_ERROR_DMA;
        fgb->dma_errors++;
    }
    if (status & FGRAB_STATUS_BUS_ERR) {
        this_pf->error |= FGRAB_ERROR_BUS;
        fgb->bus_errors++;
    }


    /*
     * If time has been requested record the current time and enter an event.
     */
    if (fgb->flags & FGRAB_WANT_TS) {
        int ind;
        do_gettimeofday(&ts);
        if ((ind=this_pf->event_cnt) < FGRAB_MAX_EVENT_CNT) {
            this_pf->event_cnt++;
            this_pf->event[ind].flag = 
                ( (status&FGRAB_STATUS_START) ? FGRAB_EVENT_START_FIELD : 0 ) |
                ( (status&FGRAB_STATUS_ODD)   ? FGRAB_EVENT_END_EVEN    : 0 ) |
                ( (status&FGRAB_STATUS_EVEN)  ? FGRAB_EVENT_END_ODD     : 0 ) ;
            this_pf->event[ind].ts = ts ;
        }
    }

    /*
     * If this is not the end of a field event the there is no need to do 
     * any other processing.
     */
    if ( ! (status & (FGRAB_STATUS_EVEN|FGRAB_STATUS_ODD)) )  goto end_intr; 

    /*
     * record the completion of fields
     */
    if (status & FGRAB_STATUS_EVEN) { /* even field */
        fgb->even_fields_captured++;
        fgb->mem->cur_field = 0;
        fgb->flags &= ~FGRAB_WANT_EVEN;
    }
    if (status & FGRAB_STATUS_ODD) { /* odd field */
        fgb->odd_fields_captured++;
        fgb->mem->cur_field = 1;
        fgb->flags &= ~FGRAB_WANT_ODD;
    }

    /*
     * Try to trap a missing or delayed interrupt event and shut down procesing
     */
    if (  ((status&FGRAB_STATUS_EVEN)&&(status&FGRAB_STATUS_ODD)) ||              /* both odd and even    */
          ((status&FGRAB_STATUS_EVEN)&&((fgb->flags&FGRAB_EXPECT_FIELD)==1)) ||   /* expect odd got even  */
          ((status&FGRAB_STATUS_ODD )&&((fgb->flags&FGRAB_EXPECT_FIELD)==0)) ) {  /* expect even got odd  */
        abort_intr(fgb,ts);
        goto end_intr;
    }
    if ((fgb->flags&FGRAB_FIELDS_MASK)==0) fgb->flags ^= FGRAB_EXPECT_FIELD ; /* toggle expected field  */

    /*
     * Does the field just completed finish the buffer ?
     */
    if (!(fgb->flags & FGRAB_WANT_MASK)                                 || 
        ((fgb->flags & FGRAB_FIELDS_MASK)==FGRAB_ALTERNATE_FIELDS)      ||
        ((fgb->flags & FGRAB_FIELDS_MASK)==FGRAB_NON_INTERLACE_FIELDS)  ) {

        fgb->frames_captured++;
        fgb->mem->frames_captured++;
        /*
         * If we are in a synch-wait state then do not close out the buffer as
         * it is still to be written to.  We increment counters above to record
         * the missed frame.
         */
        if (!fgb->synch_wait) {
            /*
             * post the completion time. 
             * Note that this is the first element of the per-frame structure
             * so that legacy meteor code should still work. RWF 25JUN97
             */
            if(fgb->flags & FGRAB_WANT_TS) {
                this_pf->frame_end = ts ;
            }
            /*
             * Record this frame's index in per-frame area
             */
            this_pf->index = fgb->frames_captured ;
            /*
             * Mark this buffer as having data and not being used by driver
             */
            this_pf->status = 1 ;
            /*
             * This used to be after update so that it pointed to the buffer being filled.
             * Now it points to the most recent buffer that is complete. RWF 25JUN97
             */
            mm->cur_frame = fgb->cur_frame;
            /*
             * This is for high-water low-water mode of frame control - RWF 25jul97
             * Mark the current frame as active.  It is up to
             * the user to clear this, but we will clear it
             * for the user for the current frame being captured
             * if we are within the water marks (see below).
             */
            mm->active |= 1 << (fgb->cur_frame - 1);
        }
        /*
         * Reset the want flags if in continuous or
         * synchronous capture mode and got both fields already.
         */
        if ((fgb->flags & (FGRAB_CONTIN|FGRAB_SYNCAP)) && !(fgb->flags & FGRAB_WANT_MASK)) {
            switch(fgb->flags & FGRAB_FIELDS_MASK) {
            case FGRAB_ONLY_ODD_FIELDS:
                fgb->flags |= FGRAB_WANT_ODD;
                break;
            case FGRAB_ONLY_EVEN_FIELDS:
                fgb->flags |= FGRAB_WANT_EVEN;
                break;
            default:
                fgb->flags |= FGRAB_WANT_MASK;
                break;
            }
        }
        /*
         * Wake up the user in single capture mode.
         */
        if(fgb->flags & FGRAB_SINGLE)
            wake_up(&fgb->waitq);
        /*
         * If the user requested to be notified via signal,
         * let them know the frame is complete.
         */
        if(fgb->proc && fgb->signal)
            send_sig(fgb->signal, fgb->proc, 1);
    
    } else {
        /* not complete frame yet, but check if user wants a signal each field anyway */
        if (fgb->sigmode==FGRAB_SIG_FIELD && fgb->proc && fgb->signal)
            send_sig(fgb->signal, fgb->proc, 1);
    }
    /*
     * Advance curent so that it points to the frame that is about to be grabbed
     * (or may actually already be in the process of being grabbed if this interrup
     * serviece routine got delayed enough).  Note that next frame really indexes
     * the buffer for the next field which may be the same as the current field.
     */
    fgb->cur_frame = fgb->next_frame;

    /*
     * If the field currently being captured completes the buffer it is working
     * on then see if we can point to buffer beyond that.
     */
    if (((fgb->flags & FGRAB_WANT_EVEN)==0)                             || 
        ((fgb->flags & FGRAB_WANT_ODD)==0)                              || 
        ((fgb->flags & FGRAB_FIELDS_MASK)==FGRAB_ALTERNATE_FIELDS)      ||
        ((fgb->flags & FGRAB_FIELDS_MASK)==FGRAB_NON_INTERLACE_FIELDS)  ) {
        unsigned next_buf_flg = 0;
        unsigned next_buffer;
        struct fgrab_per_frame *next_pf;

        /*
         *  Point to next candidate buffer in case we advance and so that
         *  we can check to see that it is free to use
         */
        next_buffer = (fgb->cur_frame == fgb->frames) ? 0 : fgb->cur_frame ;
        next_pf = (struct fgrab_per_frame *)(fgb->frame_size + 
                   (u_long)fgb->frame_buffer + fgb->frame_offset[next_buffer]) ;

        /*
         * Special handling for synchronous capture mode.
         */
        if(fgb->flags & FGRAB_SYNCAP) {
            struct fgrab_mem *mm = fgb->mem;
            /*
             * Since the user can muck with these values, we need
             * to check and see if they are sane. If they don't
             * pass the sanity check, disable the capture mode.
             * This is rather rude, but then so was the user.
             *
             * Do we really need all of this or should we just
             * eliminate the possiblity of allowing the
             * user to change hi and lo water marks while it
             * is running? XXX
             */
            /*
             * IMHO allowing these to change while running is asking for trouble.
             * Certainly the current sanity check falls apart for a multiple processor
             * system.  As a minimum need to copy lowat and hiwat before checking. RWF 25JUN97
             */
            if(mm->lowat >= fgb->frames ||
               mm->hiwat >= fgb->frames ||
               mm->lowat > mm->hiwat ) {
                fgb->progs->stop_capture(fgb->minor);
                fgb->flags &= ~(FGRAB_SYNCAP|FGRAB_WANT_MASK);
                printk("Stopped capture because of illegitimate fgrab_mem struct\n");
                printk("lowat=%d, hiwat=%d, frames=%d\n", mm->lowat, mm->hiwat, fgb->frames);
            } else {
                /*
                 * Ok, they are sane, now we want to
                 * check the water marks.
                 */
                if (mm->lowat==0 && mm->hiwat==0) 
                    fgb->synch_wait = 0;
                /*
                 * Check the alternate handshake flags
                 */
                else if (mm->lowat==-1 && mm->hiwat==-1) 
                    fgb->synch_wait = (next_pf->status == 0) ? 0 : 1 ;
                else {
                    if (mm->num_active_bufs <= mm->lowat)
                        fgb->synch_wait = 0;
                    if (mm->num_active_bufs >= mm->hiwat)
                        fgb->synch_wait = 1;
                }
                /*
                 * Clear the active frame bit for this frame
                 * and advance the counters if we are within
                 * the banks of the water marks. 
                 */
                if (!fgb->synch_wait) {
                    next_buf_flg = 1 ;
                }
            } 
        }

        /* 
         * continous ONCE mode: advance current frame, check if we got all
         */
        if ( fgb->flags & FGRAB_ONCE ) {
            /* fortran alert: cur_frame starts at 1 */
            if ( fgb->cur_frame < fgb->frames ) {
                next_buf_flg = 1;
            }
        }
        /*
         * If we need to advance do it here
         */
        if (next_buf_flg) {
            fgb->next_frame = next_buffer + 1 ;
            mm->active &= ~(1 << (next_buffer));
            init_per_frame( next_pf );
        }
    }


    /*
     * Here we set up the DMA registers for the field after the one currently being grabbed
     */
    if (status & FGRAB_STATUS_EVEN) { /* even field */
        int err;
        cli();      /* check and set of dma must be atomic wrt interrupts RWF 13aug97       */
        err=fgb->progs->set_dma_e(fgb->minor,
                                  fgb->dma_add_e, fgb->dma_str_e,
                                  (u_long)fgb->frame_buffer + fgb->frame_offset[ fgb->next_frame-1 ],
                                  fgb->frame_size);
        sti();
        if (err != 0) abort_intr(fgb,ts);
    } else
    if (status & FGRAB_STATUS_ODD) { /* odd field */
        int err;
        cli();      /* check and set of dma must be atomic wrt interrupts RWF 13aug97       */
        err=fgb->progs->set_dma_o(fgb->minor,
                                  fgb->dma_add_o, fgb->dma_str_o,
                                  (u_long)fgb->frame_buffer + fgb->frame_offset[ fgb->next_frame-1 ],
                                  fgb->frame_size);
        sti();
        if (err != 0) abort_intr(fgb,ts);
    }

end_intr:
    FGRAB_EVENT_LOG_EPROC(fgb->minor);
    return;
}

static void
abort_intr(struct fgrab *fgb, struct timeval ts) {
    u_long this_frame;
    struct fgrab_per_frame *this_pf;
    FGRAB_EVENT_LOG_STOP;
    fgb->flags &= ~FGRAB_ENABLE_INTR ;
    fgb->progs->disable_intr(fgb->minor);
    stop_capture(fgb);
    this_frame = fgb->cur_frame-1 ;
    this_pf = (struct fgrab_per_frame *)(fgb->frame_size + 
                   (u_long)fgb->frame_buffer + fgb->frame_offset[this_frame]) ;
    this_pf->error |= FGRAB_ERROR_INTR ;
    if(fgb->flags & FGRAB_WANT_TS) {
        this_pf->frame_end = ts ;
    }
    this_pf->index = fgb->frames_captured ;
    this_pf->status = 1;    /* flag this frame as done - error condition tells user the story       */
    /* wake up single frame process and send signal in conditions where user asked for it           */
    if(fgb->flags & FGRAB_SINGLE)
        wake_up(&fgb->waitq);
    if(fgb->proc && fgb->signal)
        send_sig(fgb->signal, fgb->proc, 1);
}   


static void init_per_frame(struct fgrab_per_frame *pf) {
    pf->status = 0 ;            /* flag buffer as empty         */
    pf->event_cnt = 0 ;         /* flag no events yet           */
    pf->error = 0 ;             /* start with no errors         */
}

/*---------------------------------------------------------
**
**	Fgrab character device driver routines
**
**---------------------------------------------------------
*/





static int bsd_fgrab_ioctl(struct fgrab *fgb, unsigned int cmd, void * arg) {
    int	error;  
    unsigned int	temp;
    u_long arg_val;
    u_short dir = _IOC_DIR(cmd);
    u_short size = _IOC_SIZE(cmd);
    u_short cr_cols=0, cr_rows=0;
    
    error = 0;
    
    if ( ((dir&_IOC_WRITE)!=0) && (size!=0) && (size <= sizeof(u_long)) ) {
        arg_val = 0;
        if ( (error = read_fs(arg, &arg_val, size)) ) {
            return error;
        }
    }
    
    switch (cmd) {
        
    case FGRABSTS:
        if(arg_val)
            fgb->flags |= FGRAB_WANT_TS;
        else
            fgb->flags &= ~FGRAB_WANT_TS;
        break;
    case FGRABGTS:
        if(fgb->flags & FGRAB_WANT_TS)
            return write_fs_val(1, arg, size);
        else
            return write_fs_val(0, arg, size);
        break;
    case FGRABSFPS:
        if (fgb->progs->set_fps(fgb->minor, (u_short)arg_val) < 0)
            return -EINVAL;
        else fgb->fps = arg_val ;
        break;
    case FGRABGFPS:
        return write_fs_val(fgb->fps, arg, size);
    case FGRABSSIGNAL:
        fgb->signal = (int)(arg_val&(~FGRAB_SIG_MODE_MASK));
        fgb->sigmode = (int)(arg_val&(FGRAB_SIG_MODE_MASK));
        fgb->proc = current;
        fgb->pid = current->pid;
        break;
    case FGRABGSIGNAL:
        return write_fs_val(fgb->signal, arg, size);
        
#define SETPROG(name)                                                         \
        if (fgb->progs->set_##name!=NULL)                                     \
            fgb->progs->set_##name(fgb->minor, arg_val) ;                     \
        else return(EINVAL);                                                  \
        break;
#define GETCPROG(name)                                                        \
        if (fgb->progs->get_##name!=NULL)                                     \
        return(write_fs_val((u_long)(fgb->progs->get_##name(fgb->minor)),     \
                            arg,size));                                       \
        else return(EINVAL);
            
    case FGRABSHUE:	/* set hue */
        SETPROG(hue)
    case FGRABGHUE:	/* get hue */
        GETCPROG(hue)
    case FGRABSCHCV:	/* set chrominance gain */
        SETPROG(chroma_gain)
    case FGRABGCHCV:	/* get chrominance gain */
        GETCPROG(chroma_gain)
    case FGRABSBRIG:	/* set brightness */
        SETPROG(brightness)
    case FGRABGBRIG:	/* get brightness */
        GETCPROG(brightness)
    case FGRABSCSAT:	/* set chroma saturation */
        SETPROG(chroma_saturation)
    case FGRABGCSAT:	/* get chroma saturation */
        GETCPROG(chroma_saturation)
    case FGRABSCONT:	/* set contrast */
        SETPROG(contrast)
    case FGRABGCONT:	/* get contrast */
        GETCPROG(contrast)
        
    case FGRABSHWS:	/* set horizontal window start  */
        SETPROG(hor_wndw_strt)
    case FGRABGHWS:	/* get horizontal window start  */
        GETCPROG(hor_wndw_strt)
    case FGRABSVWS:	/* set vertical window start    */
        SETPROG(vert_wndw_strt)
    case FGRABGVWS:	/* get vertical window start    */
        GETCPROG(vert_wndw_strt)
    case FGRABSPCLOCK:	/* set pixel clock              */
        SETPROG(pixel_clock)
    case FGRABGPCLOCK:	/* get pixel clock              */
        GETCPROG(pixel_clock)
            
#define SETPROGNB(name)                                                 \
                if (fgb->progs->set_##name!=NULL)                       \
                    error = fgb->progs->set_##name(fgb->minor) ;        \
                else return(-EINVAL);
                    
    case FGRABSINPUT:	/* set input device */
        switch(arg_val & FGRAB_INPUT_DEV_MASK) {
        case 0: /* default */
        case FGRAB_INPUT_DEV0:
            SETPROGNB(input_dev_0);
            fgb->flags = (fgb->flags & ~FGRAB_DEV_MASK)
                | FGRAB_DEV0;
            break;
        case FGRAB_INPUT_DEV1:
            SETPROGNB(input_dev_1);
            fgb->flags = (fgb->flags & ~FGRAB_DEV_MASK)
                | FGRAB_DEV1;
            break;
        case FGRAB_INPUT_DEV2:
            SETPROGNB(input_dev_2);
            fgb->flags = (fgb->flags & ~FGRAB_DEV_MASK)
                | FGRAB_DEV2;
            break;
        case FGRAB_INPUT_DEV3:
            SETPROGNB(input_dev_3);
            fgb->flags = (fgb->flags & ~FGRAB_DEV_MASK)
                | FGRAB_DEV3;
            break;
        case FGRAB_INPUT_DEV_SVIDEO:
            SETPROGNB(input_dev_svideo);
            fgb->flags = (fgb->flags & ~FGRAB_DEV_MASK)
                | FGRAB_DEV_SVIDEO;
            break;
        case FGRAB_INPUT_DEV_RGB:
            SETPROGNB(input_dev_rgb)
                fgb->flags = (fgb->flags & ~FGRAB_DEV_MASK)
                    | FGRAB_DEV_RGB;
            break;
        default:
            return (-EINVAL);
        }
        if (fgb->progs->set_input_ext_sync!=NULL) {
            fgb->progs->set_input_ext_sync(fgb->minor,
                                           (arg_val&FGRAB_INPUT_EXT_SYNC) ? 1 : 0 );
            fgb->flags = (fgb->flags & ~FGRAB_EXT_SYNC)
                | ((arg_val&FGRAB_INPUT_EXT_SYNC) ? FGRAB_EXT_SYNC : 0) ;
        } else {
            if (arg_val&FGRAB_INPUT_EXT_SYNC) return(-EINVAL) ;
        }
    break;
    case FGRABGINPUT:	/* get input device */
        return write_fs_val((u_long)(fgb->flags & (FGRAB_DEV_MASK|FGRAB_EXT_SYNC)), arg, size);
        
    case FGRABSFMT:	/* set input format */
        switch(arg_val & FGRAB_FORM_MASK ) {
        case 0:			/* default */
        case FGRAB_FMT_NTSC:
            SETPROGNB(fmt_ntsc);
            fgb->flags = (fgb->flags & ~FGRAB_FORM_MASK) |
                FGRAB_NTSC;
            break;
        case FGRAB_FMT_PAL:
            SETPROGNB(fmt_pal);
            fgb->flags = (fgb->flags & ~FGRAB_FORM_MASK) |
                FGRAB_PAL;
            break;
        case FGRAB_FMT_SECAM:
            SETPROGNB(fmt_secam);
            fgb->flags = (fgb->flags & ~FGRAB_FORM_MASK) |
                FGRAB_SECAM;
            break;
        case FGRAB_FMT_AUTOMODE:
            SETPROGNB(fmt_auto);
            fgb->flags = (fgb->flags & ~FGRAB_FORM_MASK) |
                FGRAB_AUTOMODE;
            break;
        default:
            return EINVAL;
        }
        break;
    case FGRABGFMT:	/* get input format */
        return write_fs_val((u_long)(fgb->flags & FGRAB_FORM_MASK), arg, size);
        
    case FGRABCAPTUR:
        temp = fgb->flags;
        switch ((int)arg_val) {
        case FGRAB_CAP_SINGLE:  /* XXX: Should merge with fgrab_read()*/
      
            if (fgb->frame_buffer==NULL) /* no frame buffer allocated */
                return(-ENOMEM);

            if (temp & FGRAB_CAP_MASK)
                return(-EIO);   /* already capturing */

            start_capture(fgb, FGRAB_SINGLE);

            /* wait for capture to complete */
            interruptible_sleep_on(&fgb->waitq);
            if (current->signal & ~current->blocked) {
                stop_capture(fgb);
                return -EINTR;
            }
            fgb->flags &= ~(FGRAB_SINGLE|FGRAB_WANT_MASK);
            break;
        case FGRAB_CAP_CONTINOUS:
            if (fgb->frame_buffer==NULL) /* no frame buffer allocated */
                return(-ENOMEM);
            if (temp & FGRAB_CAP_MASK)
                return(-EIO);   /* already capturing */
            start_capture(fgb, FGRAB_CONTIN);
            break;
        case FGRAB_CAP_CONT_ONCE:
            if (fgb->frame_buffer==NULL) /* no frame buffer allocated */
                return(-ENOMEM);
            if (temp & FGRAB_CAP_MASK)
                return(-EIO);   /* already capturing */
            start_capture(fgb, FGRAB_CONTIN|FGRAB_ONCE);
            break;
        case FGRAB_CAP_STOP_CONT:
            if (fgb->flags & FGRAB_CONTIN) {
                /* turn off capture */
                fgb->progs->stop_capture(fgb->minor);
                fgb->flags &= ~(FGRAB_ONCE|FGRAB_CONTIN|FGRAB_WANT_MASK);
            }
            break;
        default:
            error = -EINVAL;
        }
        break;
    case FGRABCAPFRM:
    {
        struct fgrab_capframe loc_frame, *frame=&loc_frame;

        if ( (error=(read_fs(arg, frame, size))) )
            return(error);
        switch (frame->command) {
        case FGRAB_CAP_N_FRAMES:
	{
            struct fgrab_mem *mem;
            if (fgb->flags & FGRAB_CAP_MASK)
                return(-EIO);
            if (fgb->frame_buffer == NULL)
                return(-ENOMEM);
            if ((fgb->frames < 2) ||
                (frame->lowat >= fgb->frames) ||
                (frame->hiwat >= fgb->frames) ||
                (frame->lowat > frame->hiwat)) 
                return(-EINVAL);
            /* fgrab_mem structure is on the page after the data */
            mem = fgb->mem; 
            fgb->cur_frame = 1;
            fgb->synch_wait = 0;
            mem->num_bufs = fgb->frames;
            mem->frame_size = fgb->frame_size;
            /* user and kernel change these */ 
            mem->lowat = frame->lowat;
            mem->hiwat = frame->hiwat;
            mem->cur_frame = 1;
            mem->active = 0;
            mem->num_active_bufs = 0;
            /* Start capture */
            start_capture(fgb, FGRAB_SYNCAP);
            break;
	}
        case FGRAB_CAP_STOP_FRAMES:
            if (fgb->flags & FGRAB_SYNCAP) {
                /* turn off capture */
                fgb->progs->stop_capture(fgb->minor);
                fgb->flags &= ~(FGRAB_SYNCAP|FGRAB_WANT_MASK);
            }
            break;
        case FGRAB_HALT_N_FRAMES:
            if(fgb->flags & FGRAB_SYNCAP) {
                fgb->progs->stop_capture(fgb->minor);
                fgb->flags &= ~(FGRAB_WANT_MASK);
            }
            break;
        case FGRAB_CONT_N_FRAMES:
            if(!(fgb->flags & FGRAB_SYNCAP)) {
                error = -EINVAL;
                break;
            }
            start_capture(fgb, FGRAB_SYNCAP);
            break;
        default:
            error = -EINVAL;
            break;
        }
        break;
    } 
    case FGRABSETGEO:
    {
        struct fgrab_geomet loc_geo, *geo=&loc_geo;
        
        error  = read_fs(arg, geo, size);
        if (error)
            return (error);
        
        /* can't change parameters while capturing */
        if (fgb->flags & FGRAB_CAP_MASK)
            return(-EBUSY);
        
        /* set/clear even/odd flags */
        switch(geo->oformat & FGRAB_GEO_FIELD_MASK) {
        case FGRAB_GEO_ODD_ONLY:
            fgb->flags = (fgb->flags & ~FGRAB_FIELDS_MASK) 
                | FGRAB_ONLY_ODD_FIELDS ;
            if ((error=fgb->progs->set_field_odd_only(fgb->minor))) 
                return(error);
            break;
        case FGRAB_GEO_EVEN_ONLY:
            fgb->flags = (fgb->flags & ~FGRAB_FIELDS_MASK) 
                | FGRAB_ONLY_EVEN_FIELDS ;
            if ((error=fgb->progs->set_field_even_only(fgb->minor))) 
                return(error);
            break;
        case FGRAB_GEO_INTERLACE:
            fgb->flags = (fgb->flags & ~FGRAB_FIELDS_MASK) 
                | FGRAB_INTERLACE_FIELDS ;
            if ((error=fgb->progs->set_field_interlace(fgb->minor))) 
                return(error);
            break;
        case FGRAB_GEO_ALTERNATE:
            fgb->flags = (fgb->flags & ~FGRAB_FIELDS_MASK) 
                | FGRAB_ALTERNATE_FIELDS ;
            if ((error=fgb->progs->set_field_alternate(fgb->minor))) 
                return(error);
            break;
        case FGRAB_GEO_NON_INTERLACE:
            fgb->flags = (fgb->flags & ~FGRAB_FIELDS_MASK) 
                | FGRAB_NON_INTERLACE_FIELDS ;
            if ((error=fgb->progs->set_field_non_interlace(fgb->minor))) 
                return(error);
            break;
        }
        

#if 0
        if (geo->frames > 32) {
            PRINTK((KERN_INFO "fgrab%d: ioctl: too many frames.\n", fgb->unit));
            error = -EINVAL;
        }
#endif

        fgb->rows = geo->rows;
        fgb->cols = geo->columns;
        fgb->frames = geo->frames;
        
        /* Tell the lewer level code what frame size and cnt we are using     */
        if ((error=fgb->progs->set_frame(fgb->minor,
                                         fgb->rows, fgb->cols, fgb->frames))) {
            return(error);
        }
        
        
        /* set defaults and end of buffer locations */
        fgb->dma_add_e[0] = 0;	/* DMA 1 even    */
        fgb->dma_add_e[1] = 0;	/* DMA 2 even    */
        fgb->dma_add_e[2] = 0;	/* DMA 3 even    */
        fgb->dma_add_o[0] = 0;	/* DMA 1 odd     */
        fgb->dma_add_o[1] = 0;	/* DMA 2 odd	 */
        fgb->dma_add_o[2] = 0;	/* DMA 3 odd     */
        fgb->dma_str_e[0] = 0;	/* Stride 1 even */
        fgb->dma_str_e[1] = 0;	/* Stride 2 even */
        fgb->dma_str_e[2] = 0;	/* Stride 3 even */
        fgb->dma_str_o[0] = 0;	/* Stride 1 odd  */
        fgb->dma_str_o[1] = 0;	/* Stride 2 odd  */
        fgb->dma_str_o[2] = 0;	/* Stride 3 odd  */
        
        switch (geo->oformat & FGRAB_GEO_OUTPUT_MASK) {

        case 0:			/* default */
        case FGRAB_GEO_RGB16:
            if (fgb->progs->set_geo_rgb16 == NULL) return(-EINVAL);
            fgb->depth = 16;
            fgb->frame_size = (fgb->rows * fgb->cols * fgb->depth) >> 3;
            fgb->flags &= ~FGRAB_OUTPUT_FMT_MASK;
            fgb->flags |= FGRAB_RGB16;
            temp = (fgb->cols * fgb->depth) >> 3;

            if ( (fgb->flags&FGRAB_FIELDS_MASK) == FGRAB_INTERLACE_FIELDS) {
                fgb->dma_add_o[0] = temp;
                fgb->dma_str_e[0] = fgb->dma_str_o[0] = temp;
            }
            if ((error=fgb->progs->set_geo_rgb16(fgb->minor))) return(error);
            break;

        case FGRAB_GEO_RGB24:
            if (fgb->progs->set_geo_rgb24 == NULL) return(-EINVAL);
            fgb->depth = 32;
            fgb->frame_size = (fgb->rows * fgb->cols * fgb->depth) >> 3;
            fgb->flags &= ~FGRAB_OUTPUT_FMT_MASK;
            fgb->flags |= FGRAB_RGB24;
            temp = (fgb->cols * fgb->depth) >> 3;

            if ( (fgb->flags&FGRAB_FIELDS_MASK) == FGRAB_INTERLACE_FIELDS) {
                fgb->dma_add_o[0] = (fgb->cols * fgb->depth) >> 3;
                fgb->dma_str_e[0] = fgb->dma_str_o[0] = 
                    (fgb->cols * fgb->depth) >> 3;
            }
            if ((error=fgb->progs->set_geo_rgb24(fgb->minor))) return(error);
            break;

        case FGRAB_GEO_MONO8:
            if (fgb->progs->set_geo_mono8 == NULL) return(-EINVAL);
            fgb->depth = 8;
            fgb->frame_size = (fgb->rows * fgb->cols * fgb->depth) >> 3;
            fgb->flags &= ~FGRAB_OUTPUT_FMT_MASK;
            fgb->flags |= FGRAB_MONO8;
            temp = (fgb->cols * fgb->depth) >> 3;

            if ( (fgb->flags&FGRAB_FIELDS_MASK) == FGRAB_INTERLACE_FIELDS) {
                fgb->dma_add_o[0] = temp;
                fgb->dma_str_e[0] = fgb->dma_str_o[0] = temp;
            }
            if ((error=fgb->progs->set_geo_mono8(fgb->minor))) return(error);
            break;

#define SETPROGNV(name)                                                       \
            if (fgb->progs->set_##name!=NULL) {                               \
                if ((error=fgb->progs->set_##name(fgb->minor)))               \
                    return(error) ;                                           \
            } else { return(-EINVAL); }
            /* These are the planar modes                                    */
        case FGRAB_GEO_YUV_9:
        case FGRAB_GEO_YUV_12:
        case FGRAB_GEO_YUV_16:
        /* These are here for compatibility with legacy code written for original meteor driver */
        case FGRAB_GEO_YUV_PLANAR:
        case FGRAB_GEO_YUV_PLANAR|FGRAB_GEO_YUV_9:          /* compatibility        */
        case FGRAB_GEO_YUV_PLANAR|FGRAB_GEO_YUV_12:         /* compatibility        */
            switch(geo->oformat & FGRAB_GEO_OUTPUT_MASK) {
            case FGRAB_GEO_YUV_9:
            case FGRAB_GEO_YUV_PLANAR|FGRAB_GEO_YUV_9:      /* compatibility        */
                SETPROGNV(geo_yuv_9)
                    fgb->depth = 9;
                cr_cols = fgb->cols >> 2;
                cr_rows = fgb->rows >> 2;
            case FGRAB_GEO_YUV_12:
            case FGRAB_GEO_YUV_PLANAR|FGRAB_GEO_YUV_12:     /* compatibility        */
                SETPROGNV(geo_yuv_12)
                    fgb->depth = 12;
                cr_cols = fgb->cols >> 1;
                cr_rows = fgb->rows >> 1;
            case FGRAB_GEO_YUV_16:
            case FGRAB_GEO_YUV_PLANAR:                  /* compatibility        */
                SETPROGNV(geo_yuv_16)
                fgb->depth = 16;
                cr_cols = fgb->cols >> 1;
                cr_rows = fgb->rows;
            }
            temp = fgb->rows * fgb->cols; /* compute (Y) frame size */
            fgb->frame_size = (temp * fgb->depth) >> 3;

            switch(fgb->flags & FGRAB_FIELDS_MASK) {
            case FGRAB_ONLY_ODD_FIELDS:
                fgb->dma_add_o[1] = temp; /* U Odd */
                fgb->dma_add_o[2] = temp + cr_cols * cr_rows; /* V Odd */
                break;
            case FGRAB_ONLY_EVEN_FIELDS:
                fgb->dma_add_e[1] = temp; /* U Even */
                fgb->dma_add_e[2] = temp + cr_cols * cr_rows; /* V Even */
                break;
            case FGRAB_ALTERNATE_FIELDS:
                fgb->dma_add_o[1] = temp; /* U Odd */
                fgb->dma_add_o[2] = temp + cr_cols * cr_rows; /* V Odd */
                fgb->dma_add_e[1] = temp; /* U Even */
                fgb->dma_add_e[2] = temp + cr_cols * cr_rows; /* V Even */
                break;
            case FGRAB_INTERLACE_FIELDS:
                /* interlaced odd even */
                fgb->dma_add_o[0] = fgb->cols; /* Y Odd */
                fgb->dma_add_e[1] = temp; /* U Even */
                fgb->dma_add_o[1] = temp + cr_cols; /* U Odd */
                fgb->dma_add_e[2] = temp + cr_cols * cr_rows; /* V Even */
                fgb->dma_add_o[2] = temp + cr_cols * cr_rows + cr_cols; /* V Odd */
                fgb->dma_str_e[0] = fgb->dma_str_o[0] = fgb->cols; /* Y Stride */
                fgb->dma_str_e[1] = fgb->dma_str_o[1] = cr_cols; /* U Stride */
                fgb->dma_str_e[2] = fgb->dma_str_o[2] = cr_cols; /* V Stride */
                break;
            }
            break;

        case FGRAB_GEO_YUV_PACKED:
            SETPROGNV(geo_yuv_packed)
            fgb->depth = 16;
            fgb->frame_size = (fgb->rows * fgb->cols * fgb->depth) >> 3;
            fgb->flags = (fgb->flags & ~FGRAB_OUTPUT_FMT_MASK) 
                | FGRAB_YUV_PACKED;
	
            if ( (fgb->flags&FGRAB_FIELDS_MASK)==FGRAB_INTERLACE_FIELDS ) {
                fgb->dma_add_o[0] = (fgb->cols * fgb->depth) >> 3;
                fgb->dma_str_e[0] = fgb->dma_str_o[0] = (fgb->cols * fgb->depth) >> 3;
            }
            break;
        default:
            error = -EINVAL;	/* invalid argument */
            PRINTK((KERN_INFO "fgrab%d: ioctl: invalid output format\n",fgb->minor));
            break;
        }
        
        /* only now can we allocate the frame buffer memory */
        /* since now the frame size is known */
        temp= fgb->frame_size * fgb->frames;
        if ( temp ) {
            /* fgrab_mem structure for SYNC Capture */
            error = alloc_frame_buffer(fgb);
        }
        if(error) return error;
        
        /* seting the dma registers with the FIRST frame or field is done in start_capture */
        
        /* set signed/unsigned chrominance */
        if (fgb->progs->set_chroma_signed != NULL) {
            fgb->progs->set_chroma_signed(fgb->minor,
                                          ((geo->oformat&FGRAB_GEO_UNSIGNED)?0:1));
        }
        break;
    }
    case FGRABGETGEO:
    {
        struct fgrab_geomet loc_geo, *geo=&loc_geo;
        geo->rows = fgb->rows;
        geo->columns = fgb->cols;
        geo->frames = fgb->frames;
        geo->oformat = 0;
        switch(fgb->flags&FGRAB_FIELDS_MASK) {
        case FGRAB_ONLY_ODD_FIELDS:
            geo->oformat |= FGRAB_GEO_ODD_ONLY ;
            break;
        case FGRAB_ONLY_EVEN_FIELDS:
            geo->oformat |= FGRAB_GEO_EVEN_ONLY ;
            break;
        case FGRAB_INTERLACE_FIELDS:
            geo->oformat |= FGRAB_GEO_INTERLACE ;
            break;
        case FGRAB_ALTERNATE_FIELDS:
            geo->oformat |= FGRAB_GEO_ALTERNATE ;
            break;
        }
        switch(fgb->flags&FGRAB_OUTPUT_FMT_MASK) {
        case FGRAB_RGB16:
            geo->oformat |= FGRAB_GEO_RGB16 ;
            break;
        case FGRAB_RGB24:
            geo->oformat |= FGRAB_GEO_RGB24 ;
            break;
        case FGRAB_MONO8:
            geo->oformat |= FGRAB_GEO_MONO8 ;
            break;
        case FGRAB_YUV_PACKED:
            geo->oformat |= FGRAB_GEO_YUV_PACKED ;
            break;
        case FGRAB_YUV_9:
            geo->oformat |= FGRAB_GEO_YUV_9 ;
            break;
        case FGRAB_YUV_12:
            geo->oformat |= FGRAB_GEO_YUV_12 ;
            break;
        case FGRAB_YUV_16:
            geo->oformat |= FGRAB_GEO_YUV_16 ;
            break;
        }
        return write_fs_ptr(geo, arg, size);
    }
    case FGRABSCOUNT:	/* (re)set error counts */
    {
        struct fgrab_counts loc_cnt, *cnt = &loc_cnt;
        if ( (error=(read_fs(arg, cnt, size))) )
            return(error);
        fgb->fifo_errors = cnt->fifo_errors;
        fgb->dma_errors = cnt->dma_errors;
        fgb->bus_errors = cnt->bus_errors;
        fgb->frames_captured = cnt->frames_captured;
        fgb->even_fields_captured = cnt->even_fields_captured;
        fgb->odd_fields_captured = cnt->odd_fields_captured;
        break;
    }
    case FGRABGCOUNT:	/* get error counts */
    {
        struct fgrab_counts loc_cnt, *cnt = &loc_cnt;

        cnt->fifo_errors = fgb->fifo_errors;
        cnt->dma_errors = fgb->dma_errors;
        cnt->bus_errors = fgb->bus_errors;
        cnt->frames_captured = fgb->frames_captured;
        cnt->even_fields_captured = fgb->even_fields_captured;
        cnt->odd_fields_captured = fgb->odd_fields_captured;
        return write_fs_ptr(cnt, arg, size);
    }
    case FGRABGFROFF:	/* get frame offsets */
    {
        struct fgrab_frame_offset off, *poff = &off; 
        int nfr = fgb->frames;
        int i;
        if (size != sizeof(struct fgrab_frame_offset)) return(-EINVAL);
        for (i=0;i<nfr;i++) off.frame_offset[i] = fgb->frame_offset[i] ;
        off.fb_size = fgb->fb_size;
        off.frame_size = fgb->frame_size;
        off.mem_off = (u_long)fgb->mem - (u_long)fgb->frame_buffer;
        return write_fs_ptr(poff, arg, size);
    }
    case FGRABCKCAP:    /* check capabilities of this board     */
    {
        struct fgrab_capabilities fgc;
        if (size != sizeof(fgc)) return(-EINVAL) ;
        error  = read_fs(arg, &fgc, size);
        if (error)
            return (error);
        switch(fgc.ioctl) {
        /*
         * The following are present for all devices
         */
        case FGRABSTS:
        case FGRABGTS:
        case FGRABSFPS:
        case FGRABGFPS:
        case FGRABSSIGNAL:
        case FGRABGSIGNAL:
        case FGRABCAPTUR:
        case FGRABCAPFRM:
        case FGRABGETGEO:
        case FGRABGFMT:
        case FGRABGINPUT:
        case FGRABSCOUNT:
        case FGRABGCOUNT:
        case FGRABGFROFF:
            break;
#define CKOPT(option,routine)                                                   \
        if ( (fgc.function&(option)) && (fgb->progs->routine == NULL) )   \
            error = -ENOTTY;
       /*
        * following are present for all devices
        * but have options that vary from device to device
        */
        case FGRABSFMT:
            CKOPT(FGRAB_FMT_NTSC,set_fmt_ntsc)
            CKOPT(FGRAB_FMT_PAL,set_fmt_pal)
            CKOPT(FGRAB_FMT_SECAM,set_fmt_secam)
            CKOPT(FGRAB_FMT_AUTOMODE,set_fmt_auto)
            break;
        case FGRABSINPUT:
            CKOPT(FGRAB_INPUT_DEV0,set_input_dev_0)
            CKOPT(FGRAB_INPUT_DEV1,set_input_dev_1)
            CKOPT(FGRAB_INPUT_DEV2,set_input_dev_2)
            CKOPT(FGRAB_INPUT_DEV3,set_input_dev_3)
            CKOPT(FGRAB_INPUT_DEV_SVIDEO,set_input_dev_svideo)
            CKOPT(FGRAB_INPUT_DEV_RGB,set_input_dev_rgb)
            CKOPT(FGRAB_INPUT_EXT_SYNC,set_input_ext_sync)
            break;
        case FGRABSETGEO:
            CKOPT(FGRAB_GEO_RGB16,set_geo_rgb16)
            CKOPT(FGRAB_GEO_RGB24,set_geo_rgb24)
            CKOPT(FGRAB_GEO_MONO8,set_geo_mono8)
            CKOPT(FGRAB_GEO_YUV_9,set_geo_yuv_9)
            CKOPT(FGRAB_GEO_YUV_12,set_geo_yuv_12)
            CKOPT(FGRAB_GEO_YUV_16,set_geo_yuv_16)
            if ( ((fgc.function&(FGRAB_GEO_YUV_PLANAR|FGRAB_GEO_YUV_9|FGRAB_GEO_YUV_12))==FGRAB_GEO_YUV_PLANAR) 
                && (fgb->progs->set_geo_yuv_16 == NULL) ) error = -ENOTTY;
            CKOPT(FGRAB_GEO_YUV_PACKED,set_geo_yuv_packed)
            CKOPT(FGRAB_GEO_UNSIGNED,set_chroma_signed)
            break;
#define CKSETPROG(name)                                                         \
        if (fgb->progs->set_##name==NULL) error = -ENOTTY;                      \
        break;
#define CKGETPROG(name)                                                        \
        if (fgb->progs->get_##name==NULL) error = -ENOTTY;                      \
        break;
        /*
         * The following ioctls exist only for some devices
         */
        case FGRABSHUE:
            CKSETPROG(hue)
        case FGRABGHUE:
            CKGETPROG(hue)
        case FGRABSCHCV:
            CKSETPROG(chroma_gain)
        case FGRABGCHCV:
            CKGETPROG(chroma_gain)
        case FGRABSBRIG:
            CKSETPROG(brightness)
        case FGRABGBRIG:
            CKGETPROG(brightness)
        case FGRABSCSAT:
            CKSETPROG(chroma_saturation)
        case FGRABGCSAT:
            CKGETPROG(chroma_saturation)
        case FGRABSCONT:
            CKSETPROG(contrast)
        case FGRABGCONT:
            CKGETPROG(contrast)
        case FGRABSHWS:
            CKSETPROG(hor_wndw_strt)
        case FGRABGHWS:
            CKGETPROG(hor_wndw_strt)
        case FGRABSVWS:
            CKSETPROG(vert_wndw_strt)
        case FGRABGVWS:
            CKGETPROG(vert_wndw_strt)
        case FGRABSPCLOCK:
            CKSETPROG(pixel_clock)
        case FGRABGPCLOCK:
            CKGETPROG(pixel_clock)
        default:
            if (fgb->progs->ckioctl != NULL) {
                error = fgb->progs->ckioctl(fgb->minor, fgc.ioctl, fgc.function);
            } else {
                error = -ENOTTY;
            }
            break;
        }
        break;
    }
    default:
        if (fgb->progs->ioctl != NULL) {
            int write_flag;
            error = fgb->progs->ioctl(fgb->minor, cmd, &arg_val,&write_flag);
            if (error) return(error);
            if (write_flag) {
                return(write_fs_val(arg_val,arg,size));
            }
        } else {
            PRINTK((KERN_INFO "fgrab%d: ioctl: invalid ioctl request\n", fgb->minor));
            error = -ENOTTY;
        }
        break;
    }  
    return(error);
}

#ifdef    FGRAB_EVENT_LOG
/*
 * This markpoint set of routines is to aid in debugging interrupt handling.
 * It keeps a circulating buffer of various events in the interrupt chain.
 * The idea is that after a problem has been detected a history of what
 * happened prior to the problem becoming evident will be available.
 */
#define FGRAB_EVENT_LOG_CNT 100
#define FGRAB_EVENT_LOG_USE_TSC
#ifdef    FGRAB_EVENT_LOG_USE_TSC
#   define FGRAB_GET_TIME(arg) get_tsc(arg)
#   define FGRAB_TIME_DEF struct tsc_time
#else  /* FGRAB_EVENT_LOG_USE_TSC */
#   define FGRAB_GET_TIME(arg) do_gettimeofday(arg)
#   define FGRAB_TIME_DEF struct timeval
#endif /* FGRAB_EVENT_LOG_USE_TSC */
struct tsc_time { unsigned long lo, hi; } ;
static void get_tsc(struct tsc_time *tsc);
static unsigned longdiv(struct tsc_time *tsc,unsigned div) ;
static struct fgrab_event_log {
    FGRAB_TIME_DEF ts;
    unsigned event;
    unsigned data;
} fel[FGRAB_EVENT_LOG_CNT];
static unsigned fel_ind;
static unsigned fel_active=0;
static FGRAB_TIME_DEF log_ref_ts;
static void
init_event_log() {
    fel_active = 0;
    FGRAB_GET_TIME(&log_ref_ts);
    for (fel_ind=0;fel_ind<FGRAB_EVENT_LOG_CNT;fel_ind++) {
        fel[fel_ind].event = FGRAB_EVENT_LOG_NULL_;
    }
    fel_ind = 0;
    fel_active = 1;
}
static void
record_event(unsigned event, unsigned data) {
    FGRAB_TIME_DEF ts;
    struct fgrab_event_log *fl;
    cli();
    if (!fel_active) {
        sti();
        return;
    }
    FGRAB_GET_TIME(&ts);
    fl = &fel[fel_ind];
    fl->ts = ts;
    fl->event = event;
    fl->data = data;
    fel_ind++;
    if (fel_ind==FGRAB_EVENT_LOG_CNT) fel_ind = 0;
    sti();
}
static void
stop_record(void) {
    fel_active = 0;
}
static void
print_event_log() {
    unsigned ind;
    fel_active = 0;
    for (ind=0;ind<FGRAB_EVENT_LOG_CNT;ind++) {
        struct fgrab_event_log *fl;
        fl = &fel[ind + fel_ind - ( (ind+fel_ind<FGRAB_EVENT_LOG_CNT) ? 0 : FGRAB_EVENT_LOG_CNT )];
        if (fl->event != FGRAB_EVENT_LOG_NULL_) {
            int d_sec, d_usec;
            char *str;
            switch(fl->event) {
            case FGRAB_EVENT_LOG_INTR_:
                str = "INTR ";
                break;
            case FGRAB_EVENT_LOG_PROC_:
                str = "PROC ";
                break;
            case FGRAB_EVENT_LOG_EPROC_:
                str = "EPROC";
                break;
            default:
                str = "INVALID";
                break;
            }
#ifdef    FGRAB_EVENT_LOG_USE_TSC
            {
                FGRAB_TIME_DEF del;
                unsigned rem;
                unsigned cpumhz=200;
                del.hi = fl->ts.hi - log_ref_ts.hi ;
                del.lo = fl->ts.lo - log_ref_ts.lo ;
                if (fl->ts.lo<log_ref_ts.lo) del.hi-- ;
                rem = longdiv(&del,cpumhz);     /* convert to microseconds      */
                if (rem<cpumhz>>1) {            /* round to nearest microsecond */
                    del.lo++;
                    if (del.lo==0) del.hi++;
                }
                d_usec = longdiv(&del,1000000); /* convert to sec - usec        */
                d_sec = del.lo ;
            }
#else  /* FGRAB_EVENT_LOG_USE_TSC */
            d_sec = fl->ts.tv_sec-log_ref_ts.tv_sec ;
            d_usec = fl->ts.tv_usec-log_ref_ts.tv_usec ;
            if (d_usec<0) {
                d_sec --;
                d_usec += 1000000;
            }
#endif /* FGRAB_EVENT_LOG_USE_TSC */
            PRINTK((KERN_INFO "fgrab: lag %2d event %s data %3d 0x%2x at %6d.%06d\n",FGRAB_EVENT_LOG_CNT-1-ind,str,fl->data,fl->data,d_sec,d_usec));
        }
    }
}
static void
get_tsc(struct tsc_time *tsc) {
    __asm__ __volatile__ (
                          "rdtsc"
                          : "=a" (tsc->lo), 
                            "=d" (tsc->hi)
                          :
                          : "%eax", "%edx");
}
#endif /* FGRAB_EVENT_LOG */
static unsigned
longdiv(struct tsc_time *tsc,unsigned div) {
    unsigned rem;
    asm (
             "movl %5,%%ecx     ;
              movl %3,%%eax     ;
              xorl %%edx,%%edx  ;
              divl %%ecx        ;
              movl %%eax,%0     ;
              movl %4,%%eax     ;
              divl %%ecx        ;
              movl %%eax,%1     ;
              movl %%edx,%2     ;"
             : "=m" (tsc->hi)  ,"=m" (tsc->lo), "=m" (rem)
             : "m" (tsc->hi)    ,"m" (tsc->lo)  , "m" (div)
             : "%eax", "%ecx", "%edx"
             );
    return(rem);
}

/*
 * Following is from drivers/scsi/hosts.c of 2.0.18.
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
