/*    fgrab_pci.c  -- PCI specific code for generic frame grabber
 *    
 *    See file COPYING for copyright info
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/bios32.h>
#include <linux/pci.h>
#include "fgrab.h"
#define FGRAB_PCI_INTERNAL
#include "fgrab_pci.h"
#undef  FGRAB_PCI_INTERNAL
/*
 * Check and reset pciStatus errors. 
 */

static struct fgrab_pci *fgrab_pci[MAXFGRABS+1] ;

static void setup(struct fgrab_pci *pci,int irq, unsigned minor) ;
static int pci_detect(int (*dev_register)(int irq, void *data, 
                                          struct fgrab_pci *pci),
                      int (*dev_signature)(void *data),
                      void *data,
                      unsigned int pci_vendor_id,unsigned int pci_device_id,
                      unsigned int window_size) ;
static void unwind(unsigned minor) ;
static unsigned check_bus(unsigned minor) ;
static void enable_dma(unsigned minor) ;
static void disable_dma(unsigned minor) ;
static void set_latency(struct fgrab_pci *pci) ;
static void *get_addr(unsigned minor) ;


struct pci_progs fgrab_pci_progs = { setup, pci_detect, unwind, check_bus, 
                                     enable_dma, disable_dma, get_addr }  ;

static struct symbol_table fgrab_pci_syms = {
#include <linux/symtab_begin.h>
    X(fgrab_pci_progs),
#include <linux/symtab_end.h>
} ;    

/* Designed to be used as a module */
#ifdef MODULE

void cleanup_module(void) {
}


int init_module(void) {
    int minor;
    register_symtab(&fgrab_pci_syms);
    for (minor=0;minor<MAXFGRABS;minor++) fgrab_pci[minor] = NULL ;
    return(0);
}
#endif /*MODULE*/


static void
setup(struct fgrab_pci *pci,int irq, unsigned minor) {
    fgrab_pci[minor] = pci ;
    pci->minor = minor ;
}

static void unwind(unsigned minor) {
    vfree((void *)fgrab_pci[minor]->addr); /* Unwind vremap() */
    fgrab_pci[minor] = NULL ;
    }


static unsigned 
check_bus(unsigned minor) {
    struct fgrab_pci *pci = fgrab_pci[minor] ;
    u_short status;
    register u_short bus = pci->bus, devFun = pci->dev_fn;

    pcibios_read_config_word(bus, devFun, PCI_STATUS, &status);
    /* Clear errors */
    pcibios_write_config_word(bus, devFun, PCI_STATUS, status); 
    return((unsigned)status);
}

static void
enable_dma(unsigned minor) {
    struct fgrab_pci *pci = fgrab_pci[minor] ;
    u_short cmd;
    register u_short bus = pci->bus, devFun = pci->dev_fn;

    pcibios_read_config_word(bus, devFun, PCI_COMMAND, &cmd);
    cmd |= PCI_COMMAND_MASTER ;
    pcibios_write_config_word(bus, devFun, PCI_COMMAND, cmd); 
}

static void
disable_dma(unsigned minor) {
    struct fgrab_pci *pci = fgrab_pci[minor] ;
    u_short cmd;
    register u_short bus = pci->bus, devFun = pci->dev_fn;

    pcibios_read_config_word(bus, devFun, PCI_COMMAND, &cmd);
    cmd &= ~PCI_COMMAND_MASTER ;
    pcibios_write_config_word(bus, devFun, PCI_COMMAND, cmd); 
}

static int pci_detect(int (*dev_register)(int irq, void *data, 
                                          struct fgrab_pci *pci),
                      int (*dev_signature)(void *data),
                      void *data,
                      unsigned int pci_vendor_id,unsigned int pci_device_id,
                      unsigned int window_size) {
    u_short pci_index;
    u_char irq;
    int pci_ret;
    unsigned indices[MAXFGRABS],minor;

    PRINTK((KERN_INFO "fgrab_pci: Checking for vendor %d device 0%o\n",
            pci_vendor_id,pci_device_id));

  
    if ( ! (pci_ret=pcibios_present()) ) {
        printk(KERN_ERR "fgrab_pci: bios_present error %d: %s\n",
                pci_ret,pcibios_strerror(pci_ret));
        return -EUNATCH; /* Let's get creative */
    }   

    /*
     * First find out which pci indices already have entries for this
     * device
     */
    for (pci_index=0;pci_index<MAXFGRABS;pci_index++) indices[pci_index]=0;
    for (minor=0;minor<MAXFGRABS;minor++) {
        if (fgrab_pci[minor] == NULL) continue;
        if ( (fgrab_pci[minor]->pci_vendor_id != pci_vendor_id) ||
             (fgrab_pci[minor]->pci_device_id != pci_device_id) ) continue;
        indices[minor] = 1 ;
    }

    for ( pci_index = 0; pci_index < MAXFGRABS; pci_index++ ) {
        u_char bus, devFun;
        struct fgrab_pci pci;
        unsigned int pci_addr;
        unsigned short command;
        int bioserr ;

        if (indices[pci_index] != 0) continue;

        pci_ret = pcibios_find_device(pci_vendor_id,pci_device_id,
                                      pci_index, &bus, &devFun);
        if (pci_ret != PCIBIOS_SUCCESSFUL) {
            PRINTK((KERN_INFO "fgrab_pci: find_device returned %d at index %d: %s\n",
                   pci_ret,pci_index,pcibios_strerror(pci_ret)));
            break;
        }
        pcibios_read_config_byte(bus, devFun, PCI_INTERRUPT_LINE, &irq);
        pcibios_read_config_dword(bus, devFun, PCI_BASE_ADDRESS_0, &pci_addr);
        pcibios_read_config_word(bus, devFun,  PCI_COMMAND, &command);
        command |= (PCI_COMMAND_MEMORY);
        if ((bioserr = pcibios_write_config_word(bus, devFun, PCI_COMMAND, command))) {
            PRINTK((KERN_INFO "fgrab: pcibios write failed (errno=%d) pci_index=%d.\n",
                    bioserr,pci_index));
            return(-EFAULT);
        }

        pci.index = pci_index ;
        pci.bus = bus;
        pci.dev_fn = devFun;
        pci.pci_vendor_id = pci_vendor_id ;
        pci.pci_device_id = pci_device_id ;
        set_latency(&pci);
        {  /* Map device regs into our address space */
            u_long pci_page = pci_addr & PAGE_MASK;
            u_long pci_off = pci_addr & ~PAGE_MASK;
            u_long bus_addr = pci_addr;
            pci_page = (u_long)vremap((u_long)pci_page, window_size);
            if ( ! pci_page ) {
                PRINTK((KERN_INFO "fgrab: vremap failed pci_index=%d.\n",pci_index));
                return(-EFAULT) ;
            }
            pci_addr = pci_off | pci_page ;
            PRINTK((KERN_INFO "fgrab: pci_index=%d bus address 0x%lx mapped to 0x%x\n",
                                                 pci_index,bus_addr,pci_addr));
        }
        pci.addr = pci_addr ;
        fgrab_pci[FGRAB_MINOR_SETUP] = &pci;
        /* Need to check that this is really the device and not just another
           PCI device with same PCI signature                               */
        if (!dev_signature(data)) {
            PRINTK((KERN_INFO "fgrab_pci: device signature at index %d does not match\n",pci_index));
            vfree((void *)pci.addr); /* Unwind vremap() */
            continue ;
        }
        /* dev_register assigns us a minor device number, irq, and allocates 
           memory both for info the generic code needs and for info we need
           at this level.  It also provides a mechanism for us to tell it
           what subroutines are associated with this device.                */
        if (dev_register(irq,data,&pci)) {
            printk(KERN_ERR "fgrab: pci: cannot register pci_index %d\n",
                   pci_index);
            return(-ENOMEM);    /* FIXME need to track error returns        */
        }
        
    }
return(0);
}

static void *get_addr(unsigned minor) {
    return((void *)fgrab_pci[minor]->addr);
}

static void set_latency(struct fgrab_pci *pci) { 
    /* Adjust latency */
    u_short bus = pci->bus, dev_fn = pci->dev_fn;
    u_char latency;

    pcibios_read_config_byte(bus, dev_fn, PCI_LATENCY_TIMER, &latency);
    latency = FGRAB_LATENCY;
    pcibios_write_config_byte(bus, dev_fn, PCI_LATENCY_TIMER, latency);
#ifdef DEBUG_FGRAB
    pcibios_read_config_byte(bus, dev_fn, PCI_LATENCY_TIMER, &latency);
    PRINTK((KERN_INFO "fgrab_pci: Index %d Latency = %d\n", pci->index, latency));
#endif
}

/*
 * Following is from drivers/scsi/hosts.c of 2.0.18.
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
