/*
 *      saa7116.c  --  Routines specific to the saa7116 chip.
 *    
 *    See file COPYING for copyright info
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/delay.h>
#define SAA7116_INTERNAL
#include "fgrab_pci.h"
#include "saa7116.h"
#undef  SAA7116_INTERNAL
#include "fgrab.h"

/* This array has one pointer for each board plus one to use durring setup  */
static struct saa7116 *saa7116[MAXFGRABS+1];

/* declarations of routines accessible through saa7116_progs                */
static int  detect(int (*dev_register)(unsigned irq,struct saa7116 *saa7116),
                   int (*dev_signature)(void)) ;
static void unwind(unsigned minor) ;
static void setup(struct saa7116 *saa7116,int irq, unsigned minor) ;
static int  set_dma_e(unsigned minor,
                      unsigned long *off_e, unsigned long *stride_e,
                      unsigned long base, unsigned long fb_size) ;
static int  set_dma_o(unsigned minor,
                      unsigned long *off_o, unsigned long *stride_o,
                      unsigned long base, unsigned long fb_size) ;
static void start_capture(unsigned minor, unsigned single_flag) ;
static void stop_capture(unsigned minor) ;
static int  check_intr(unsigned minor) ;
static void disable_intr(unsigned minor) ;
static void enable_intr(unsigned minor) ;
static int  check_err(unsigned minor) ;
static void clear_err(unsigned minor) ;
static void zero_regs(unsigned minor) ;
static void set_capt_ctl(unsigned minor, unsigned val) ;
static void set_i2c_clocks(unsigned minor, 
                           unsigned phase_4, unsigned phase_3, 
                           unsigned phase_2, unsigned phase_1) ;
static void set_fifo_trigger(unsigned minor, unsigned flag) ;
static void enable_field_toggle(unsigned minor) ;
static void disable_field_toggle(unsigned minor) ;
static void reset_int_flags(unsigned minor) ;
static void set_fifo_limits(unsigned minor) ;
static void probe_range(unsigned minor) ;
static void set_field_masks(unsigned minor, 
                            unsigned field_mask_e,unsigned length_e,
                            unsigned field_mask_o,unsigned length_o) ;
static void set_field_odd_only(unsigned minor) ;
static void set_field_even_only(unsigned minor) ;
static void set_field_both(unsigned minor) ;
static void set_geo_rgb16(unsigned minor) ;
static void set_geo_rgb24(unsigned minor) ;
static void set_geo_yuv_9(unsigned minor) ;
static void set_geo_yuv_12(unsigned minor) ;
static void set_geo_yuv_16(unsigned minor) ;
static void set_geo_yuv_packed(unsigned minor) ;
static unsigned get_capt_ctl(unsigned minor) ;
static int get_i2c_data(unsigned minor) ;
static int talk_i2c(unsigned minor, u_char dev, u_char addr, u_char data) ;



/* Filling in the address of programs accessible via saa7116_progs      */
struct saa7116_progs saa7116_progs = {
    detect, unwind, setup, set_dma_e, set_dma_o, 
    start_capture, stop_capture,
    check_intr, enable_intr, disable_intr, check_err, clear_err,
    zero_regs, set_capt_ctl, set_i2c_clocks, set_fifo_trigger, 
    enable_field_toggle, disable_field_toggle, 
    reset_int_flags, set_fifo_limits,
    probe_range, set_field_masks,
    set_field_odd_only, set_field_even_only, set_field_both,
    set_geo_rgb16, set_geo_rgb24, 
    set_geo_yuv_9, set_geo_yuv_12, set_geo_yuv_16, set_geo_yuv_packed,
    get_capt_ctl, get_i2c_data, talk_i2c
} ;

static struct symbol_table saa7116_syms = {
#include <linux/symtab_begin.h>
    X(saa7116_progs),
#include <linux/symtab_end.h>
} ;    

/* Designed to be used as a module */
#ifdef MODULE

void cleanup_module(void) {
}


int init_module(void) {
    register_symtab(&saa7116_syms);
    return(0);
}
#endif /*MODULE*/


static void
setup(struct saa7116 *saa,int irq, unsigned minor) {
    saa7116[minor] = saa ;
    saa->minor = minor ;
    fgrab_pci_progs.setup(&(saa->pci),irq,minor) ;
}

struct saa7116_detect { 
    int (*dev_register)(unsigned irq,struct saa7116 *saa7116); 
    int (*dev_signature)(void); 
} ;
static int saa7116_register(int irq, void *data, struct fgrab_pci *pci) {
    int err;
    struct saa7116_detect *det=(struct saa7116_detect *)data;
    struct saa7116 saa;
    saa.pci = *pci ;
    saa.reg = (struct saa7116_reg *)fgrab_pci_progs.get_addr(FGRAB_MINOR_SETUP);
    PRINTK((KERN_INFO "fgrab: register saa7116 address 0x%x.\n", (unsigned)saa.reg));
    err=det->dev_register(irq,&saa);
    return(err);
}
static int saa7116_signature(void *data) {
    struct saa7116_detect *det=(struct saa7116_detect *)data ;
    struct saa7116 saa;
    /* we need to bootstrap this process somehow - need talki2c for sig */
    saa.reg = (struct saa7116_reg *)fgrab_pci_progs.get_addr(FGRAB_MINOR_SETUP);
    saa7116[FGRAB_MINOR_SETUP] = &saa ;
    return(det->dev_signature());
}
static int detect(int (*dev_register)(unsigned irq,struct saa7116 *saa7116),
                  int (*dev_signature)(void)) {
    struct saa7116_detect det={dev_register,dev_signature};
    return(fgrab_pci_progs.detect(saa7116_register,saa7116_signature,
                                  (void *)&det,
                                  PCI_VENDOR_ID_INTEL,PCI_DEVICE_ID_INTEL_7116,
                                  sizeof(struct saa7116_reg)));
}

static void unwind(unsigned minor) {
    fgrab_pci_progs.unwind(minor);
}

/*
 * Set the capture control register (CSR1)
 */
static void set_capt_ctl(unsigned minor, unsigned val) {
    register struct saa7116_reg *reg = saa7116[minor]->reg ;
    reg->capt_ctl = val;
}

static void set_i2c_clocks(unsigned minor, 
                           unsigned phase_4, unsigned phase_3, 
                           unsigned phase_2, unsigned phase_1) {
    register struct saa7116_reg *reg = saa7116[minor]->reg ;
    reg->i2c_clocks = ((phase_4 & (0xff))<<24) | ((phase_3 & (0xff))<<16) |
                      ((phase_2 & (0xff))<<8) | ((phase_1 & (0xff))<<0) ;
}

static void 
set_fifo_trigger(unsigned minor, unsigned flag) {
    register struct saa7116_reg *reg = saa7116[minor]->reg ;
    reg->fifo_trigger = ((struct ff_trig){ packed: (flag&SAA7116_PACKED) ? SAA7116_FIFO_TRIGGER : 0 , 
                                           planar: (flag&SAA7116_PLANAR) ? SAA7116_FIFO_TRIGGER : 0 });
}

static void 
enable_field_toggle(unsigned minor) {
    register struct saa7116_reg *reg = saa7116[minor]->reg ;
    reg->field_toggle = 0x00000107;	      /* enable field toggle, etc. */
}    

static void 
disable_field_toggle(unsigned minor) {
    register struct saa7116_reg *reg = saa7116[minor]->reg ;
    reg->field_toggle = 0x00000103;	      /* disable field toggle, etc. */
}    

static void 
reset_int_flags(unsigned minor) {
    register struct saa7116_reg *reg = saa7116[minor]->reg ;
    reg->ints_ctl = SAA7116_INTR_ST_MASK;	      /* Disable all interrupts and Clear status */
}    

static void
disable_intr(unsigned minor)
{
    saa7116[minor]->reg->capt_ctl &= 0x8ff0;	           /* disable future captures           */
    saa7116[minor]->reg->ints_ctl = SAA7116_INTR_ST_MASK ; /* disable interrupts, clr status    */
}

static void
enable_intr(unsigned minor) {
    /* 
     * Enable all interrupts (incl start of frame) and clear status
     */
    saa7116[minor]->reg->ints_ctl = SAA7116_INTR_ST_MASK | SAA7116_INTR_EN_MASK ;
}

static void set_fifo_limits(unsigned minor) {
    register struct saa7116_reg *reg = saa7116[minor]->reg ;
    reg->fifo_limits = ((SAA7116_FIFO_ALMOST_EMPTY)<<16) | SAA7116_FIFO_ALMOST_FULL ;
}    

static void probe_range(unsigned minor) {
    register struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    int maxRange;

    /*
     * Determine the size of the dma_end registers. It appears there
     * may be different saa7116 versions floating around, and that more
     * bits may be added in future (according to Philips docs).
     */
    s7116->dma_end_e = s7116->dma_end_o = ~0;
    maxRange = (s7116->dma_end_e & s7116->dma_end_o) | 0xff;
    PRINTK((KERN_INFO "fgrab%d: maxRange %x\n", minor, maxRange));
    s7116->dma_end_e = s7116->dma_end_o = 0;
}

/*
 * Talk to a device on the i2c (or iic) bus.
 */

static int talk_i2c(unsigned minor, u_char dev, u_char addr, u_char data) {
    struct saa7116_reg *reg=saa7116[minor]->reg ;
    int jifcnt ;
    /* execute i2c cycle */
    reg->i2c_cmd = 
        ((struct i2c_cmd){w_data:data, reg_addr: addr, dev_addr:dev, xfer:1});

    /* and wait for completion */
    udelay(70); /* This is from DT315[25] documentation RWF     */
    /* Loop is legacy from meteor driver RWF        */
    for (jifcnt=0;;jifcnt++) {  
        if ( ! reg->i2c_cmd.xfer ) 
            break;
        if (jifcnt >= SAA7116_WAIT_JIFFS) {
            goto error;
        }
        fgrab_wait_jiffies(1);
    }

    if ( reg->i2c_stat.d_abort ) {
        reg->i2c_stat.d_abort = 1; /* Reset bit to zero */
        goto error;
    }
    return 0;
error:
    return 1;
}

static int get_i2c_data(unsigned minor) {
    return(saa7116[minor]->reg->i2c_stat.r_data) ;
}


/*
 * Zero the SAA7116 registers
 */
static void zero_regs(unsigned minor) {
    struct saa7116_reg *reg = saa7116[minor]->reg ;
    /* Clear and reset everything  */
    memset(reg, 0, sizeof(struct saa7116_reg)); 
}

/*
 * There is a handshaking between the even and odd dma set
 * routines to try to trap when we shouldn't write to dma registers.
 * The even dma routine checks that it is not beyond the end of the
 * odd field.  If it is still withing bounds of odd field then 
 * it assumes that the even dma registers will be set before the 
 * start of even field.  If it is not withing bounds it does not
 * set the dma registers and returns a 1. This routine should be run
 * with interrupts disabled to be sure no body can sneak in between
 * the check and the setting of the dma registers.  The even dma
 * routine also resets the bit for even field done so that the
 * odd dma routine can use that as an indicator that it shouldn't
 * try setting the odd dma registers.  The odd dma routine does 
 * the same thing where odd and even are interchanged.
 */
static int
set_dma_e(unsigned minor,
          unsigned long *off_e, unsigned long *stride_e,
          unsigned long base, unsigned long fb_size)
{
    struct saa7116_reg *s7116 = saa7116[minor]->reg; 
    unsigned captctl;
    captctl=s7116->capt_ctl;
    if ( (captctl&SAA7116_CSR1_FLD_DN_ODD) != 0) {
        return(1);  /* flag error condition */
    }
    s7116->capt_ctl = captctl | SAA7116_CSR1_FLD_DN_EVE ;
    s7116->dma_e[0] = off_e[0] + base;
    s7116->dma_e[1] = off_e[1] + base;
    s7116->dma_e[2] = off_e[2] + base;
    s7116->stride_e[0] = stride_e[0];
    s7116->stride_e[1] = stride_e[1];
    s7116->stride_e[2] = stride_e[2];
    s7116->dma_end_e = (base + fb_size - 1 );	 
    return(0);
}

static int
set_dma_o(unsigned minor,
          unsigned long *off_o, unsigned long *stride_o,
          unsigned long base, unsigned long fb_size)
{
    struct saa7116_reg *s7116 = saa7116[minor]->reg; 
    unsigned captctl;
    captctl=s7116->capt_ctl;
    if ( (captctl&SAA7116_CSR1_FLD_DN_EVE) != 0) {
        return(1);  /* flag error condition */
    }
    s7116->capt_ctl = captctl | SAA7116_CSR1_FLD_DN_ODD ;
    s7116->dma_o[0] = off_o[0] + base;
    s7116->dma_o[1] = off_o[1] + base;
    s7116->dma_o[2] = off_o[2] + base;
    s7116->stride_o[0] = stride_o[0];
    s7116->stride_o[1] = stride_o[1];
    s7116->stride_o[2] = stride_o[2];
    s7116->dma_end_o = (base + fb_size - 1 );	 
    return(0);
}

static void set_field_masks(unsigned minor, 
                            unsigned field_mask_e,unsigned length_e,
                            unsigned field_mask_o,unsigned length_o) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    s7116->field_mask_e = field_mask_e ;
    s7116->field_mask_o = field_mask_o ;
    s7116->mask_lengths = ( (length_o<<16) | (length_e) );
}

static void set_field_odd_only(unsigned minor) {
    saa7116[minor]->flags = (saa7116[minor]->flags & ~SAA7116_FIELDS_MASK)
                             | SAA7116_ONLY_ODD_FIELDS ;
}
static void set_field_even_only(unsigned minor) {
    saa7116[minor]->flags = (saa7116[minor]->flags & ~SAA7116_FIELDS_MASK)
                             | SAA7116_ONLY_EVEN_FIELDS ;
}
static void set_field_both(unsigned minor) {
    saa7116[minor]->flags = (saa7116[minor]->flags & ~SAA7116_FIELDS_MASK) ;
}

static void
start_capture(unsigned minor, unsigned single_flag) {
    u_long cap;
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;

    switch(saa7116[minor]->flags) {
    case SAA7116_ONLY_EVEN_FIELDS:
        cap = (single_flag) ? 0x8ff4 : 0x8ff1 ;
        break;
    case SAA7116_ONLY_ODD_FIELDS:
        cap = (single_flag) ? 0x8ff8 : 0x8ff2 ;
        break;
    default:
        cap = (single_flag) ? 0x8ffc : 0x8ff3 ;
        break;
    }
    
    s7116->capt_ctl = cap;
}

static void
stop_capture(unsigned minor) /* Stop the capture */ {
    /*
     * Note that this does not stop the capture till end of current field RWF 28jul97
     */
    saa7116[minor]->reg->capt_ctl = 0x8ff0;
}


/* Check to see if this board needs servicing                           */
static int
check_intr(unsigned minor) {
    register volatile u_long *cap;
    register u_long status;
    register struct saa7116_reg *s7116=saa7116[minor]->reg; 

    status = s7116->ints_ctl;
    if ( ! (status & SAA7116_INTR_ST_MASK) ) {
        return(1);		/* This board does not want attention   */
    }
    /* clear acknowledged interrupt status */
    s7116->ints_ctl =  (SAA7116_INTR_EN_MASK | SAA7116_INTR_ST_MASK ) & status ;		

    cap = &(s7116->capt_ctl);
    saa7116[minor]->status = status ;
    saa7116[minor]->cap_err = *cap & SAA7116_CSR1_ERR_MASK ;
    /*
     * Clear Error.  Note that field done is used in set_dma routines
     * to determine when it is safe to change dma registers. RWF 13aug97
     */
    *cap |= SAA7116_CSR1_ERR_MASK ;
    saa7116[minor]->bus_err = fgrab_pci_progs.check_bus(minor) ;
    return(0) ;
}

#define BUS_STATUS_CHECK (                    \
            PCI_STATUS_SIG_TARGET_ABORT |     \
            PCI_STATUS_REC_TARGET_ABORT |     \
            PCI_STATUS_REC_MASTER_ABORT |     \
            PCI_STATUS_DETECTED_PARITY )
static int
check_err(unsigned minor) {
    struct saa7116 *saa = saa7116[minor];
    unsigned saastatus;
    unsigned status = 0;
    if (saa->cap_err & 0x300) {
        status |= FGRAB_STATUS_FIFO_ERR ;
    }
    if (saa->cap_err & 0xc00) {
        status |= FGRAB_STATUS_DMA_ERR ;
    }
    if (BUS_STATUS_CHECK&saa->bus_err) {
        status |= FGRAB_STATUS_BUS_ERR ;
    }

    saastatus = saa->status;
    status |= ( (saastatus & (SAA7116_FLD_END_EVE)) ? FGRAB_STATUS_EVEN   : 0 ) |
              ( (saastatus & (SAA7116_FLD_END_ODD)) ? FGRAB_STATUS_ODD    : 0 ) |
              ( (saastatus & (SAA7116_FLD_START  )) ? FGRAB_STATUS_START  : 0 ) ;
    return(status) ;
}


static void
clear_err(unsigned minor) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg; 
    /* 0x0f30 means clear error and field done */
    s7116->capt_ctl |= SAA7116_CSR1_ERR_MASK | SAA7116_CSR1_DN_MASK ;
    /* Clear interrupt status register          */
    s7116->ints_ctl =  (SAA7116_INTR_EN_MASK  & s7116->ints_ctl) |
                        SAA7116_INTR_ST_MASK ;
}

static void set_geo_rgb16(unsigned minor) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    s7116->rt_mode_e = s7116->rt_mode_o =
        (struct rt_mode){mode: 0x01, route: 0xeeeeee}; /* RGB 16 */
}


static void set_geo_rgb24(unsigned minor) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    s7116->rt_mode_e = s7116->rt_mode_o =
        (struct rt_mode){mode: 0x00, route: 0x393939}; /* RGB 24 */
}

static void set_geo_yuv_9(unsigned minor) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    s7116->rt_mode_e = s7116->rt_mode_o = 
        (struct rt_mode){mode: 0xc3, route: 0xaaaaff};
}

static void set_geo_yuv_12(unsigned minor) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    s7116->rt_mode_e = s7116->rt_mode_o = 
        (struct rt_mode){mode: 0xc2, route: 0xaaaaff};
}

static void set_geo_yuv_16(unsigned minor) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    s7116->rt_mode_e = s7116->rt_mode_o = 
        (struct rt_mode){mode: 0xc1, route: 0xaaaaff};
}

static void set_geo_yuv_packed(unsigned minor) {
    struct saa7116_reg *s7116 = saa7116[minor]->reg ;
    s7116->rt_mode_e = s7116->rt_mode_o =
        (struct rt_mode){mode: 0x41, route: 0xeeeeee}; /* YUV 16 packed */
}

static unsigned get_capt_ctl(unsigned minor) {
    return(saa7116[minor]->reg->capt_ctl) ;
}

/*
 * Following is from drivers/scsi/hosts.c of 2.0.18.
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-indent-level: 4
 * c-brace-imaginary-offset: 0
 * c-brace-offset: -4
 * c-argdecl-indent: 4
 * c-label-offset: -4
 * c-continued-statement-offset: 4
 * c-continued-brace-offset: 0
 * indent-tabs-mode: nil
 * tab-width: 8
 * End:
 */
